import os
from typing import Dict, List, Optional

import netsquid as ns
import pytest
from netqasm.lang.instr import TrappedIonFlavour, core

from qoala.lang.ehi import EhiNodeInfo, UnitModule
from qoala.lang.hostlang import BasicBlock, BasicBlockType
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.lhi import (
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiProcNodeInfo,
    LhiTopologyBuilder,
)
from qoala.runtime.ntf import GenericNtf, TrappedIonNtf
from qoala.runtime.program import ProgramInput, ProgramInstance
from qoala.runtime.task import (
    HostLocalTask,
    LocalRoutineTask,
    PostCallTask,
    PreCallTask,
    TaskGraphBuilder,
    TaskInfo,
)
from qoala.sim.build import build_network_from_lhi
from qoala.sim.driver import CpuDriver, QpuDriver, SharedSchedulerMemory
from qoala.sim.network import ProcNodeNetwork
from qoala.sim.scheduler import CpuEdfScheduler, QpuScheduler
from qoala.util.builder import ObjectBuilder
from qoala.util.logging import LogManager

CL = BasicBlockType.CL
CC = BasicBlockType.CC
QL = BasicBlockType.QL
QC = BasicBlockType.QC


def get_pure_host_program() -> QoalaProgram:
    program_text = """
META_START
    name: alice
    parameters:
    csockets:
    epr_sockets:
META_END

^b0 {type = CL}:
    var_x = assign_cval() : 3
    var_y = assign_cval() : 5
^b1 {type = CL}:
    var_z = assign_cval() : 9
    """

    return QoalaParser(program_text).parse()


def get_lr_program() -> QoalaProgram:
    program_text = """
META_START
    name: alice
    parameters:
    csockets:
    epr_sockets:
META_END

^b0 {type = CL}:
    x = assign_cval() : 3
^b1 {type = QL}:
    tuple<y> = run_subroutine(tuple<x>) : add_one

SUBROUTINE add_one
    params: x
    returns: y
    uses: 
    keeps:
    request:
  NETQASM_START
    load C0 @input[0]
    set C1 1
    add R0 C0 C1
    store R0 @output[0]
  NETQASM_END
    """

    return QoalaParser(program_text).parse()


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def load_program_trapped_ion(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text, flavour=TrappedIonFlavour()).parse()


def setup_network(internal_sched_latency: float = 0) -> ProcNodeNetwork:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=3)
    latencies = LhiLatencies(
        host_instr_time=1000,
        qnos_instr_time=2000,
        host_peer_latency=3000,
        internal_sched_latency=internal_sched_latency,
    )
    link_info = LhiLinkInfo.perfect(duration=20_000)

    alice_lhi = LhiProcNodeInfo(
        name="alice", id=0, topology=topology, latencies=latencies
    )
    nodes = {0: "alice", 1: "bob"}
    network_lhi = LhiNetworkInfo.fully_connected(nodes, link_info)
    bob_lhi = LhiProcNodeInfo(name="bob", id=1, topology=topology, latencies=latencies)
    ntfs = [GenericNtf(), GenericNtf()]
    return build_network_from_lhi([alice_lhi, bob_lhi], ntfs, network_lhi)


def setup_network_trapped_ion() -> ProcNodeNetwork:
    num_qubits = 3
    topology = LhiTopologyBuilder.trapped_ion_default_perfect_gates(
        num_qubits=num_qubits
    )
    latencies = LhiLatencies(
        host_instr_time=1000, qnos_instr_time=2000, host_peer_latency=3000
    )
    link_info = LhiLinkInfo.perfect(duration=20_000)

    alice_lhi = LhiProcNodeInfo(
        name="alice", id=0, topology=topology, latencies=latencies
    )
    nodes = {0: "alice", 1: "bob"}
    network_lhi = LhiNetworkInfo.fully_connected(nodes, link_info)
    bob_lhi = LhiProcNodeInfo(name="bob", id=1, topology=topology, latencies=latencies)
    ntfs = [TrappedIonNtf(), TrappedIonNtf()]
    return build_network_from_lhi([alice_lhi, bob_lhi], ntfs, network_lhi)


def instantiate(
    program: QoalaProgram,
    ehi: EhiNodeInfo,
    pid: int = 0,
    inputs: Optional[ProgramInput] = None,
) -> ProgramInstance:
    unit_module = UnitModule.from_full_ehi(ehi)

    if inputs is None:
        inputs = ProgramInput.empty()

    return ProgramInstance(
        pid,
        program,
        inputs,
        unit_module=unit_module,
    )


def test_cpu_scheduler():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_pure_host_program()

    pid = 0
    instance = ObjectBuilder.simple_program_instance(program, pid)

    procnode.scheduler.submit_program_instance(instance)

    tasks_with_start_times = [
        (HostLocalTask(0, 0, "b0"), 0),
        (HostLocalTask(1, 0, "b1"), 1000),
    ]
    graph = TaskGraphBuilder.linear_tasks_with_start_times(tasks_with_start_times)

    mem = SharedSchedulerMemory()
    driver = CpuDriver("alice", mem, procnode.host.processor, procnode.memmgr)
    scheduler = CpuEdfScheduler(
        "alice", 0, driver, procnode.memmgr, procnode.host.interface
    )
    scheduler.add_tasks(graph.get_tasks())

    ns.sim_reset()
    scheduler.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid).host_mem.read("var_y") == 5
    assert procnode.memmgr.get_process(pid).host_mem.read("var_z") == 9

    assert ns.sim_time() == 1000


def test_cpu_scheduler_no_time():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_pure_host_program()

    pid = 0
    instance = ObjectBuilder.simple_program_instance(program, pid)

    procnode.scheduler.submit_program_instance(instance)

    tasks = [HostLocalTask(0, 0, "b0"), HostLocalTask(1, 0, "b1")]
    tinfos: Dict[int, TaskInfo] = {
        task.task_id: TaskInfo.only_task(task) for task in tasks
    }

    mem = SharedSchedulerMemory()
    driver = CpuDriver("alice", mem, procnode.host.processor, procnode.memmgr)
    scheduler = CpuEdfScheduler(
        "alice", 0, driver, procnode.memmgr, procnode.host.interface
    )
    scheduler.add_tasks(tinfos)

    ns.sim_reset()
    scheduler.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid).host_mem.read("var_y") == 5
    assert procnode.memmgr.get_process(pid).host_mem.read("var_z") == 9

    assert ns.sim_time() == 0


def test_cpu_scheduler_2_processes():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_pure_host_program()

    pid0 = 0
    pid1 = 1
    instance0 = ObjectBuilder.simple_program_instance(program, pid0)
    instance1 = ObjectBuilder.simple_program_instance(program, pid1)

    procnode.scheduler.submit_program_instance(instance0)
    procnode.scheduler.submit_program_instance(instance1)

    tasks_with_start_times = [
        (HostLocalTask(0, pid0, "b0"), 0),
        (HostLocalTask(1, pid1, "b0"), 500),
        (HostLocalTask(2, pid0, "b1"), 1000),
        (HostLocalTask(3, pid1, "b1"), 1500),
    ]
    graph = TaskGraphBuilder.linear_tasks_with_start_times(tasks_with_start_times)

    mem = SharedSchedulerMemory()
    driver = CpuDriver("alice", mem, procnode.host.processor, procnode.memmgr)
    scheduler = CpuEdfScheduler(
        "alice", 0, driver, procnode.memmgr, procnode.host.interface
    )
    scheduler.add_tasks(graph.get_tasks())

    ns.sim_reset()
    scheduler.start()
    ns.sim_run(duration=1000)

    assert procnode.memmgr.get_process(pid0).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid0).host_mem.read("var_y") == 5
    with pytest.raises(KeyError):
        procnode.memmgr.get_process(pid0).host_mem.read("var_z")
        procnode.memmgr.get_process(pid1).host_mem.read("var_z")

    ns.sim_run()
    assert procnode.memmgr.get_process(pid0).host_mem.read("var_z") == 9
    assert procnode.memmgr.get_process(pid1).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid1).host_mem.read("var_y") == 5
    assert procnode.memmgr.get_process(pid1).host_mem.read("var_z") == 9

    assert ns.sim_time() == 1500


def test_qpu_scheduler():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_lr_program()

    pid = 0
    instance = ObjectBuilder.simple_program_instance(program, pid)

    procnode.scheduler.submit_program_instance(instance)

    shared_ptr = 0

    cpu_tasks_with_start_times = [
        (HostLocalTask(0, 0, "b0"), 0),
        (PreCallTask(1, 0, "b1", shared_ptr), 1000),
        (PostCallTask(2, 0, "b1", shared_ptr), 5000),
    ]
    cpu_graph = TaskGraphBuilder.linear_tasks_with_start_times(
        cpu_tasks_with_start_times
    )
    qpu_tasks_with_start_times = [
        (LocalRoutineTask(3, 0, "b1", shared_ptr), 2000),
    ]
    qpu_graph = TaskGraphBuilder.linear_tasks_with_start_times(
        qpu_tasks_with_start_times
    )
    qpu_graph.get_tinfo(3).ext_predecessors.add(1)
    cpu_graph.get_tinfo(2).ext_predecessors.add(3)

    mem = SharedSchedulerMemory()
    cpu_driver = CpuDriver("alice", mem, procnode.host.processor, procnode.memmgr)
    cpu_scheduler = CpuEdfScheduler(
        "alice", 0, cpu_driver, procnode.memmgr, procnode.host.interface
    )
    cpu_scheduler.add_tasks(cpu_graph.get_tasks())

    qpu_driver = QpuDriver(
        "alice",
        mem,
        procnode.qnos.processor,
        procnode.memmgr,
        procnode.memmgr,
    )
    qpu_scheduler = QpuScheduler("alice", 0, qpu_driver, procnode.memmgr, None)
    qpu_scheduler.add_tasks(qpu_graph.get_tasks())

    cpu_scheduler.set_other_scheduler(qpu_scheduler)
    qpu_scheduler.set_other_scheduler(cpu_scheduler)

    LogManager.set_log_level("INFO")
    ns.sim_reset()
    cpu_scheduler.start()
    qpu_scheduler.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid).host_mem.read("y") == 4

    assert ns.sim_time() == 5000


def test_qpu_scheduler_2_processes():
    LogManager.set_task_log_level("INFO")

    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_lr_program()

    pid0 = 0
    pid1 = 1
    instance0 = ObjectBuilder.simple_program_instance(program, pid0)
    instance1 = ObjectBuilder.simple_program_instance(program, pid1)

    procnode.scheduler.submit_program_instance(instance0)
    procnode.scheduler.submit_program_instance(instance1)

    shared_ptr_pid0 = 0
    shared_ptr_pid1 = 1

    cpu_tasks = [
        (HostLocalTask(0, pid0, "b0", CL), 0),
        (HostLocalTask(1, pid1, "b0", CL), 500),
        (PreCallTask(2, pid0, "b1", shared_ptr_pid0), 1000),
        (PreCallTask(3, pid1, "b1", shared_ptr_pid1), 1000),
        (PostCallTask(4, pid0, "b1", shared_ptr_pid0), 1000),
        (PostCallTask(5, pid1, "b1", shared_ptr_pid1), 1000),
    ]
    cpu_graph = TaskGraphBuilder.linear_tasks_with_start_times(cpu_tasks)
    qpu_tasks = [
        (LocalRoutineTask(6, pid0, "b1", shared_ptr_pid0), 1000),
        (LocalRoutineTask(7, pid1, "b1", shared_ptr_pid1), 1000),
    ]
    qpu_graph = TaskGraphBuilder.linear_tasks_with_start_times(qpu_tasks)

    cpu_graph.get_tinfo(4).ext_predecessors.add(6)
    cpu_graph.get_tinfo(5).ext_predecessors.add(7)
    qpu_graph.get_tinfo(6).ext_predecessors.add(2)
    qpu_graph.get_tinfo(7).ext_predecessors.add(3)

    mem = SharedSchedulerMemory()
    cpu_driver = CpuDriver("alice", mem, procnode.host.processor, procnode.memmgr)
    cpu_scheduler = CpuEdfScheduler(
        "alice", 0, cpu_driver, procnode.memmgr, procnode.host.interface
    )
    cpu_scheduler.add_tasks(cpu_graph.get_tasks())

    qpu_driver = QpuDriver(
        "alice",
        mem,
        procnode.qnos.processor,
        procnode.memmgr,
        procnode.memmgr,
    )
    qpu_scheduler = QpuScheduler("alice", 0, qpu_driver, procnode.memmgr, None)
    qpu_scheduler.add_tasks(qpu_graph.get_tasks())

    cpu_scheduler.set_other_scheduler(qpu_scheduler)
    qpu_scheduler.set_other_scheduler(cpu_scheduler)

    ns.sim_reset()
    cpu_scheduler.start()
    qpu_scheduler.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid0).host_mem.read("y") == 4
    assert procnode.memmgr.get_process(pid1).host_mem.read("y") == 4

    assert ns.sim_time() == 1000


def test_host_program():

    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program = load_program("test_scheduling_alice.iqoala")
    pid = 0
    instance = instantiate(program, alice.local_ehi, pid, ProgramInput({"bob_id": 1}))

    used_blocks = {"blk_host0", "blk_host1"}

    new_blocks: List[BasicBlock] = []
    for block in instance.program.blocks:
        if block.name in used_blocks:
            new_blocks.append(block)

    instance.program.blocks = new_blocks

    alice.scheduler.submit_program_instance(instance, remote_pid=0)
    bob.scheduler.submit_program_instance(instance, remote_pid=0)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 3 * alice.local_ehi.latencies.host_instr_time
    assert alice.memmgr.get_process(pid).host_mem.read("var_z") == 9
    assert bob.memmgr.get_process(pid).host_mem.read("var_z") == 9


def test_lr_program():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program = load_program("test_scheduling_alice.iqoala")
    pid = 0
    instance = instantiate(program, alice.local_ehi, pid, ProgramInput({"bob_id": 1}))

    used_blocks = {"blk_host2", "blk_add_one"}

    new_blocks = []
    for block in instance.program.blocks:
        if block.name in used_blocks:
            new_blocks.append(block)

    instance.program.blocks = new_blocks

    alice.scheduler.submit_program_instance(instance, remote_pid=0)
    bob.scheduler.submit_program_instance(instance, remote_pid=0)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    host_instr_time = alice.local_ehi.latencies.host_instr_time
    qnos_instr_time = alice.local_ehi.latencies.qnos_instr_time
    expected_duration = 3 * host_instr_time + 5 * qnos_instr_time
    assert ns.sim_time() == expected_duration
    assert alice.memmgr.get_process(pid).host_mem.read("y") == 4
    assert bob.memmgr.get_process(pid).host_mem.read("y") == 4


def test_epr_md_1():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    used_blocks_alice = {"blk_epr_md_1"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    used_blocks_bob = {"blk_epr_md_1"}
    new_blocks_bob = []
    for block in instance_bob.program.blocks:
        if block.name in used_blocks_bob:
            new_blocks_bob.append(block)

    instance_bob.program.blocks = new_blocks_bob

    alice.scheduler.submit_program_instance(instance_alice, instance_bob.pid)
    bob.scheduler.submit_program_instance(instance_bob, instance_alice.pid)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    host_instr_time = alice.local_ehi.latencies.host_instr_time
    expected_duration = alice.network_ehi.get_link(0, 1).duration
    assert ns.sim_time() == 2 * host_instr_time + expected_duration
    alice_outcome = alice.memmgr.get_process(pid).host_mem.read("m")
    bob_outcome = bob.memmgr.get_process(pid).host_mem.read("m")
    assert alice_outcome == bob_outcome


def test_epr_md_2():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    used_blocks_alice = {"blk_epr_md_2"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    used_blocks_bob = {"blk_epr_md_2"}
    new_blocks_bob = []
    for block in instance_bob.program.blocks:
        if block.name in used_blocks_bob:
            new_blocks_bob.append(block)

    instance_bob.program.blocks = new_blocks_bob

    alice.scheduler.submit_program_instance(instance_alice, instance_bob.pid)
    bob.scheduler.submit_program_instance(instance_bob, instance_alice.pid)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    host_instr_time = alice.local_ehi.latencies.host_instr_time
    expected_duration = alice.network_ehi.get_link(0, 1).duration * 2
    assert ns.sim_time() == 2 * host_instr_time + expected_duration
    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    alice_outcomes = [alice_mem.read("m0"), alice_mem.read("m1")]
    bob_outcomes = [bob_mem.read("m0"), bob_mem.read("m1")]
    assert alice_outcomes == bob_outcomes


def test_epr_ck_1():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    used_blocks_alice = {"blk_epr_ck_1", "blk_meas_q0"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    used_blocks_bob = {"blk_epr_ck_1", "blk_meas_q0"}
    new_blocks_bob = []
    for block in instance_bob.program.blocks:
        if block.name in used_blocks_bob:
            new_blocks_bob.append(block)

    instance_bob.program.blocks = new_blocks_bob

    alice.scheduler.submit_program_instance(instance_alice, instance_bob.pid)
    bob.scheduler.submit_program_instance(instance_bob, instance_alice.pid)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    # subrt meas_q0 has 3 classical instructions + 1 meas instruction
    subrt_class_time = 3 * alice.local_ehi.latencies.qnos_instr_time
    subrt_meas_time = alice.local_ehi.find_single_gate(0, core.MeasInstruction).duration
    subrt_time = subrt_class_time + subrt_meas_time
    host_instr_time = alice.local_ehi.latencies.host_instr_time
    epr_time = alice.network_ehi.get_link(0, 1).duration
    expected_duration = 4 * host_instr_time + epr_time + subrt_time
    assert ns.sim_time() == expected_duration
    alice_outcome = alice.memmgr.get_process(pid).host_mem.read("p")
    bob_outcome = bob.memmgr.get_process(pid).host_mem.read("p")
    assert alice_outcome == bob_outcome


def test_epr_ck_2():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    used_blocks_alice = {"blk_epr_ck_2", "blk_meas_q0_q1"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    used_blocks_bob = {"blk_epr_ck_2", "blk_meas_q0_q1"}
    new_blocks_bob = []
    for block in instance_bob.program.blocks:
        if block.name in used_blocks_bob:
            new_blocks_bob.append(block)

    instance_bob.program.blocks = new_blocks_bob

    alice.scheduler.submit_program_instance(instance_alice, instance_bob.pid)
    bob.scheduler.submit_program_instance(instance_bob, instance_alice.pid)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    # subrt meas_q0_q1 has 6 classical instructions + 2 meas instruction
    subrt_class_time = 6 * alice.local_ehi.latencies.qnos_instr_time
    meas_time = alice.local_ehi.find_single_gate(0, core.MeasInstruction).duration
    subrt_time = subrt_class_time + 2 * meas_time
    host_instr_time = alice.local_ehi.latencies.host_instr_time
    epr_time = alice.network_ehi.get_link(0, 1).duration
    expected_duration = 4 * host_instr_time + 2 * epr_time + subrt_time
    assert ns.sim_time() == expected_duration
    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    alice_outcomes = [alice_mem.read("p0"), alice_mem.read("p1")]
    bob_outcomes = [bob_mem.read("p0"), bob_mem.read("p1")]
    assert alice_outcomes == bob_outcomes


def test_cc():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    # TODO: add start times ?

    used_blocks_alice = {"blk_prep_cc", "blk_send", "blk_host1"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    used_blocks_bob = {"blk_prep_cc", "blk_recv", "blk_host1"}
    new_blocks_bob = []
    for block in instance_bob.program.blocks:
        if block.name in used_blocks_bob:
            new_blocks_bob.append(block)

    instance_bob.program.blocks = new_blocks_bob

    alice.scheduler.submit_program_instance(instance_alice, instance_bob.pid)
    bob.scheduler.submit_program_instance(instance_bob, instance_alice.pid)

    assert alice.local_ehi.latencies.host_peer_latency == 3000
    assert alice.local_ehi.latencies.host_instr_time == 1000

    ns.sim_reset()
    network.start()
    ns.sim_run()

    # assert ns.sim_time() == expected_duration
    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    assert bob_mem.read("msg") == 25
    assert alice_mem.read("var_z") == 9
    assert bob_mem.read("var_z") == 9


def test_full_program():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice, instance_bob.pid)
    bob.scheduler.submit_program_instance(instance_bob, instance_alice.pid)

    ns.sim_reset()
    network.start()
    ns.sim_run()


def test_jump_instruction():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_jump", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 5000  # 5 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 0
    assert alice_mem.read("var_y") == 1


def test_beq_instruction_1():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_beq_1", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 5000  # 5 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 1
    assert alice_mem.read("var_y") == 1


def test_beq_instruction_2():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_beq_2", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 7000  # 7 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 9
    assert alice_mem.read("var_y") == 9


def test_bne_instruction_1():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_bne_1", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 5000  # 5 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 2
    assert alice_mem.read("var_y") == 3


def test_bne_instruction_2():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_bne_2", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 7000  # 7 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 9
    assert alice_mem.read("var_y") == 9


def test_bgt_instruction_1():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_bgt_1", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 5000  # 5 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 5
    assert alice_mem.read("var_y") == 4


def test_bgt_instruction_2():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_bgt_2", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 7000  # 7 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 9
    assert alice_mem.read("var_y") == 9


def test_blt_instruction_1():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_blt_1", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 5000  # 5 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 6
    assert alice_mem.read("var_y") == 7


def test_blt_instruction_2():
    network = setup_network()
    alice = network.nodes["alice"]

    program_alice = load_program("test_jumping_and_branching.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    used_blocks_alice = {"blk_blt_2", "blk_temp", "blk_last"}
    new_blocks_alice = []
    for block in instance_alice.program.blocks:
        if block.name in used_blocks_alice:
            new_blocks_alice.append(block)

    instance_alice.program.blocks = new_blocks_alice

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 7000  # 7 * 1000

    alice_mem = alice.memmgr.get_process(pid).host_mem
    assert alice_mem.read("var_x") == 9
    assert alice_mem.read("var_y") == 9


def test_measure_all():
    network = setup_network_trapped_ion()
    alice = network.nodes["alice"]

    program_alice = load_program_trapped_ion("test_measure_all.iqoala")

    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()
    ns.sim_run()
    alice_outcome = alice.memmgr.get_process(pid).host_mem.read("m0")
    alice_outcome2 = alice.memmgr.get_process(pid).host_mem.read("m1")
    alice_outcome3 = alice.memmgr.get_process(pid).host_mem.read("m2")
    assert alice_outcome == alice_outcome2 == alice_outcome3 == 0


def test_internal_sched_latency():
    network = setup_network(internal_sched_latency=500)
    alice = network.nodes["alice"]

    program_alice = load_program("test_internal_sched_latency.iqoala")

    pid = 0
    instance_alice = instantiate(program_alice, alice.local_ehi, pid)

    alice.scheduler.submit_program_instance(instance_alice, 0)

    ns.sim_reset()
    assert ns.sim_time() == 0
    network.start()

    network.nodes["bob"].stop()
    ns.sim_run()

    total_host_instr_time = 4 * 1000
    total_internal_sched_latency = 2 * 500  # 2 CL blocks => 2 * 500
    total_time = total_host_instr_time + total_internal_sched_latency
    assert ns.sim_time() == total_time


if __name__ == "__main__":
    test_cpu_scheduler()
    test_cpu_scheduler_no_time()
    test_cpu_scheduler_2_processes()
    test_qpu_scheduler()
    test_qpu_scheduler_2_processes()
    test_host_program()
    test_lr_program()
    test_epr_md_1()
    test_epr_md_2()
    test_epr_ck_1()
    test_epr_ck_2()
    test_cc()
    test_full_program()
    test_jump_instruction()
    test_beq_instruction_1()
    test_beq_instruction_2()
    test_bne_instruction_1()
    test_bne_instruction_2()
    test_bgt_instruction_1()
    test_bgt_instruction_2()
    test_blt_instruction_1()
    test_blt_instruction_2()
    test_measure_all()
    test_internal_sched_latency()



import logging
from copy import deepcopy
from typing import Any, Dict, Generator, List, Optional

from pydynaa import EventExpression
from qoala.lang.request import CallbackType, EprType, QoalaRequest
from qoala.runtime.lhi import INSTR_MEASURE_INSTANT
from qoala.runtime.memory import ProgramMemory, RunningRequestRoutine
from qoala.runtime.message import Message, RrCallTuple
from qoala.sim.entdist.entdist import EntDistRequest
from qoala.sim.netstack.netstackinterface import NetstackInterface, NetstackLatencies
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice, QDeviceCommand
from qoala.sim.qnos.qnosprocessor import QnosProcessor
from qoala.util.logging import LogManager


class NetstackProcessor:
    def __init__(
        self, interface: NetstackInterface, latencies: NetstackLatencies
    ) -> None:
        self._interface = interface
        self._latencies = latencies

        self._name = f"{interface.name}_NetstackProcessor"

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

        # memory of current program, only not-None when processor is active
        self._current_prog_mem: Optional[ProgramMemory] = None

        self._current_routine: Optional[RunningRequestRoutine] = None

    def _prog_mem(self) -> ProgramMemory:
        # May only be called when processor is active
        assert self._current_prog_mem is not None
        return self._current_prog_mem

    def _routine(self) -> RunningRequestRoutine:
        # May only be called when processor is active
        assert self._current_routine is not None
        return self._current_routine

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    def _execute_entdist_request(
        self, request: EntDistRequest
    ) -> Generator[EventExpression, None, bool]:
        self._interface.send_entdist_msg(Message(-1, -1, request))
        result = yield from self._interface.receive_entdist_msg()
        self._logger.info(f"got result {result}")
        return result.content is not None

    def _allocate_for_pair(
        self, process: QoalaProcess, request: QoalaRequest, index: int
    ) -> int:
        memmgr = self._interface.memmgr
        pid = process.pid

        virt_id = request.virt_ids.get_id(index)
        memmgr.allocate(pid, virt_id)

        return virt_id

    def _create_entdist_request(
        self, process: QoalaProcess, request: QoalaRequest, virt_id: int
    ) -> EntDistRequest:
        memmgr = self._interface.memmgr
        pid = process.pid
        phys_id = memmgr.phys_id_for(pid, virt_id)

        epr_sck = process.epr_sockets[request.epr_socket_id]

        return EntDistRequest(
            local_node_id=self._interface.node_id,
            remote_node_id=request.remote_id,
            local_qubit_id=phys_id,
            local_pid=epr_sck.local_pid,
            remote_pid=epr_sck.remote_pid,
        )

    def measure_epr_qubit(
        self, process: QoalaProcess, virt_id: int
    ) -> Generator[EventExpression, None, int]:
        phys_id = self._interface.memmgr.phys_id_for(process.pid, virt_id)
        # Should have been allocated by `_handle_multi_pair_md`
        assert phys_id is not None
        # Use the special INSTR_MEASURE_INSTANT instruction so that measuring
        # doesn't take time.
        commands = [QDeviceCommand(INSTR_MEASURE_INSTANT, [phys_id])]
        m = yield from self.qdevice.execute_commands(commands=commands)
        assert isinstance(m, int)
        return m

    def _handle_multi_pair_ck(
        self, process: QoalaProcess, routine_name: str
    ) -> Generator[EventExpression, None, bool]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request
        assert request.typ == EprType.CREATE_KEEP
        num_pairs = request.num_pairs

        log_str = (
            f"Handling CK request {routine.name} "
            f"for PID {process.pid} and remot ID = {request.remote_id}"
        )
        self._logger.debug(log_str)

        virt_ids: List[int] = []
        for i in range(num_pairs):
            virt_id = self._allocate_for_pair(process, request, i)
            virt_ids.append(virt_id)
            self._logger.info(f"trying to create EPR pair {i} (virt ID = {virt_id})")
            entdist_req = self._create_entdist_request(process, request, virt_id)

            self._interface.send_entdist_msg(Message(-1, -1, entdist_req))

        for i, virt_id in enumerate(virt_ids):
            result_msg = yield from self._interface.receive_entdist_msg()
            self._logger.info(f"got result {result_msg}")
            result = result_msg.content is not None

            if result:  # success
                self._logger.info(
                    f"successfully created EPR pair {i} (virt ID = {virt_id})"
                )
            if not result:
                self._logger.info(f"failed creating EPR pair {i} (virt ID = {virt_id})")
                self._logger.info(f"freeing qubits {virt_ids}")
                for vid in virt_ids:
                    self._interface.memmgr.free(process.pid, vid)
                return False
        return True

    def _handle_multi_pair_md(
        self, process: QoalaProcess, routine_name: str
    ) -> Generator[EventExpression, None, bool]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request
        assert request.typ == EprType.MEASURE_DIRECTLY
        num_pairs = request.num_pairs

        outcomes: List[int] = []

        if routine.callback_type == CallbackType.SEQUENTIAL:
            raise NotImplementedError
        else:
            for i in range(num_pairs):
                virt_id = self._allocate_for_pair(process, request, i)
                entdist_req = self._create_entdist_request(process, request, virt_id)
                # Create EPR pair
                result = yield from self._execute_entdist_request(entdist_req)
                if result:
                    # Measure local qubit
                    m = yield from self.measure_epr_qubit(process, virt_id)
                    outcomes.append(m)
                # Free virt qubit
                self._interface.memmgr.free(process.pid, virt_id)
                if not result:
                    return False

        shared_mem = process.prog_memory.shared_mem
        results_addr = running_routine.result_addr
        shared_mem.write_rr_out(results_addr, outcomes)
        return True

    def handle_multi_pair(
        self, process: QoalaProcess, routine_name: str
    ) -> Generator[EventExpression, None, bool]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request

        if request.typ == EprType.CREATE_KEEP:
            result = yield from self._handle_multi_pair_ck(process, routine_name)
        elif request.typ == EprType.MEASURE_DIRECTLY:
            result = yield from self._handle_multi_pair_md(process, routine_name)
        else:
            raise NotImplementedError
        return result

    def handle_multi_pair_callback(
        self, process: QoalaProcess, routine_name: str, qnosprocessor: QnosProcessor
    ) -> Generator[EventExpression, None, None]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine

        assert routine.callback is not None
        cb_routine = process.get_local_routine(routine.callback)
        for virt_id in cb_routine.metadata.qubit_use:
            if self._interface.memmgr.phys_id_for(process.pid, virt_id) is None:
                self._interface.memmgr.allocate(process.pid, virt_id)

        # for WAIT_ALL, there should be at most 1 callback.
        # So we can access index 0 of the cb input/output addresses.
        yield from qnosprocessor.assign_local_routine(
            process=process,
            routine_name=routine.callback,
            input_addr=running_routine.cb_input_addrs[0],
            result_addr=running_routine.cb_output_addrs[0],
        )

        # Free CR qubits
        for virt_id in cb_routine.metadata.qubit_use:
            if virt_id not in cb_routine.metadata.qubit_keep:
                self._interface.memmgr.free(process.pid, virt_id)

    def handle_single_pair(
        self, process: QoalaProcess, routine_name: str, index: int
    ) -> Generator[EventExpression, None, bool]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request

        virt_id = self._allocate_for_pair(process, request, index)
        entdist_req = self._create_entdist_request(process, request, virt_id)
        result = yield from self._execute_entdist_request(entdist_req)
        if not result:
            self._interface.memmgr.free(process.pid, virt_id)
        return result

    def handle_single_pair_callback(
        self,
        process: QoalaProcess,
        routine_name: str,
        qnosprocessor: QnosProcessor,
        index: int,
    ) -> Generator[EventExpression, None, None]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine

        assert routine.callback is not None
        cb_routine = process.get_local_routine(routine.callback)
        for virt_id in cb_routine.metadata.qubit_use:
            if self._interface.memmgr.phys_id_for(process.pid, virt_id) is None:
                self._interface.memmgr.allocate(process.pid, virt_id)

        assert qnosprocessor is not None
        yield from qnosprocessor.assign_local_routine(
            process=process,
            routine_name=routine.callback,
            input_addr=running_routine.cb_input_addrs[index],
            result_addr=running_routine.cb_output_addrs[index],
        )

        # Free CR qubits
        for virt_id in cb_routine.metadata.qubit_use:
            if virt_id not in cb_routine.metadata.qubit_keep:
                self._interface.memmgr.free(process.pid, virt_id)

    def instantiate_routine(
        self,
        process: QoalaProcess,
        rrcall: RrCallTuple,
        args: Dict[str, Any],
    ) -> None:
        """Instantiates and activates routine."""
        routine = process.get_request_routine(rrcall.routine_name)
        instance = deepcopy(routine)
        instance.instantiate(args)

        running_routine = RunningRequestRoutine(
            instance,
            rrcall.input_addr,
            rrcall.result_addr,
            rrcall.cb_input_addrs,
            rrcall.cb_output_addrs,
        )
        process.qnos_mem.add_running_request_routine(running_routine)

    def _handle_req_routine_md(
        self, process: QoalaProcess, routine_name: str
    ) -> Generator[EventExpression, None, None]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request
        assert request.typ == EprType.MEASURE_DIRECTLY
        num_pairs = request.num_pairs

        outcomes: List[int] = []

        if routine.callback_type == CallbackType.SEQUENTIAL:
            raise NotImplementedError
        else:
            for i in range(num_pairs):
                virt_id = self._allocate_for_pair(process, request, i)
                entdist_req = self._create_entdist_request(process, request, virt_id)
                # Create EPR pair
                yield from self._execute_entdist_request(entdist_req)
                # Measure local qubit
                m = yield from self.measure_epr_qubit(process, virt_id)
                # Free virt qubit
                self._interface.memmgr.free(process.pid, virt_id)
                outcomes.append(m)

        shared_mem = process.prog_memory.shared_mem
        results_addr = running_routine.result_addr
        shared_mem.write_rr_out(results_addr, outcomes)

    def _handle_req_routine_ck(
        self,
        process: QoalaProcess,
        routine_name: str,
        qnosprocessor: Optional[QnosProcessor] = None,
    ) -> Generator[EventExpression, None, None]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request
        num_pairs = request.num_pairs

        if routine.callback_type == CallbackType.SEQUENTIAL:
            for i in range(num_pairs):
                virt_id = self._allocate_for_pair(process, request, i)
                entdist_req = self._create_entdist_request(process, request, virt_id)
                yield from self._execute_entdist_request(entdist_req)

                if routine.callback is not None:
                    # TODO: write CR inputs to shared memory

                    # Allocate qubits for CR
                    cb_routine = process.get_local_routine(routine.callback)
                    for virt_id in cb_routine.metadata.qubit_use:
                        if (
                            self._interface.memmgr.phys_id_for(process.pid, virt_id)
                            is None
                        ):
                            self._interface.memmgr.allocate(process.pid, virt_id)

                    assert qnosprocessor is not None
                    yield from qnosprocessor.assign_local_routine(
                        process=process,
                        routine_name=routine.callback,
                        input_addr=running_routine.cb_input_addrs[i],
                        result_addr=running_routine.cb_output_addrs[i],
                    )

                    # Free CR qubits
                    for virt_id in cb_routine.metadata.qubit_use:
                        if virt_id not in cb_routine.metadata.qubit_keep:
                            self._interface.memmgr.free(process.pid, virt_id)
        else:
            for i in range(num_pairs):
                virt_id = self._allocate_for_pair(process, request, i)
                entdist_req = self._create_entdist_request(process, request, virt_id)
                yield from self._execute_entdist_request(entdist_req)

            if routine.callback is not None:
                # TODO: write CR inputs to shared memory

                # Allocate qubits for CR
                cb_routine = process.get_local_routine(routine.callback)
                for virt_id in cb_routine.metadata.qubit_use:
                    if self._interface.memmgr.phys_id_for(process.pid, virt_id) is None:
                        self._interface.memmgr.allocate(process.pid, virt_id)

                # for WAIT_ALL, there should be at most 1 callback.
                # So we can access index 0 of the cb input/output addresses.

                assert qnosprocessor is not None
                yield from qnosprocessor.assign_local_routine(
                    process=process,
                    routine_name=routine.callback,
                    input_addr=running_routine.cb_input_addrs[0],
                    result_addr=running_routine.cb_output_addrs[0],
                )

                # Free CR qubits
                for virt_id in cb_routine.metadata.qubit_use:
                    if virt_id not in cb_routine.metadata.qubit_keep:
                        self._interface.memmgr.free(process.pid, virt_id)

    def assign_request_routine(
        self,
        process: QoalaProcess,
        rrcall: RrCallTuple,
        qnosprocessor: Optional[QnosProcessor] = None,
    ) -> Generator[EventExpression, None, None]:
        routine = process.get_request_routine(rrcall.routine_name)
        global_args = process.prog_instance.inputs.values
        self.instantiate_routine(process, rrcall, global_args)

        if routine.request.typ == EprType.CREATE_KEEP:
            yield from self._handle_req_routine_ck(
                process, rrcall.routine_name, qnosprocessor
            )
        elif routine.request.typ == EprType.MEASURE_DIRECTLY:
            yield from self._handle_req_routine_md(process, rrcall.routine_name)
        else:
            raise NotImplementedError



from __future__ import annotations

from typing import Generator, List

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.message import Message
from qoala.sim.componentprot import ComponentProtocol, PortListener
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.events import EVENT_WAIT, SIGNAL_NSTK_ENTD_MSG


class EntDistInterface(ComponentProtocol):
    def __init__(
        self,
        comp: EntDistComponent,
        ehi_network: EhiNetworkInfo,
    ) -> None:
        super().__init__(name=f"{comp.name}_protocol", comp=comp)
        self._comp = comp
        self._ehi_network = ehi_network

        self._all_node_names: List[str] = self._ehi_network.get_all_node_names()

        for node in self._all_node_names:
            self.add_listener(
                f"node_{node}",
                PortListener(
                    self._comp.node_in_port(node), f"{SIGNAL_NSTK_ENTD_MSG}_{node}"
                ),
            )

    def remote_id_to_peer_name(self, remote_id: int) -> str:
        return self._ehi_network.nodes[remote_id]

    def send_node_msg(self, node: str, msg: Message) -> None:
        self._comp.node_out_port(node).tx_output(msg)

    def receive_node_msg(self, node: str) -> Generator[EventExpression, None, Message]:
        yield from self._wait_for_msg(f"node_{node}", f"{SIGNAL_NSTK_ENTD_MSG}_{node}")
        return self._pop_any_msg(f"node_{node}")

    def wait_for_any_msg(self) -> Generator[EventExpression, None, None]:
        yield from self._wait_for_msg_any_source(
            [f"node_{node}" for node in self._all_node_names],
            [f"{SIGNAL_NSTK_ENTD_MSG}_{node}" for node in self._all_node_names],
        )

    def receive_msg(self) -> Generator[EventExpression, None, Message]:
        yield from self._wait_for_msg_any_source(
            [f"node_{node}" for node in self._all_node_names],
            [f"{SIGNAL_NSTK_ENTD_MSG}_{node}" for node in self._all_node_names],
        )
        return self._pop_any_msg_any_source(
            [f"node_{node}" for node in self._all_node_names]
        )

    def pop_all_messages(self) -> List[Message]:
        return self._pop_all_messages([f"node_{node}" for node in self._all_node_names])

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from __future__ import annotations

import os
from typing import List

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramInput
from qoala.sim.build import build_network_from_config
from qoala.sim.network import ProcNodeNetwork


def get_config() -> ProcNodeConfig:
    params = NvParams()
    topology = TopologyConfig.from_nv_params(num_qubits=5, params=params)
    return ProcNodeConfig(
        node_name="alice",
        node_id=0,
        topology=topology,
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("NvNtf"),
    )


def create_network(
    node_cfg: ProcNodeConfig,
) -> ProcNodeNetwork:
    network_cfg = ProcNodeNetworkConfig(nodes=[node_cfg], links=[])
    return build_network_from_config(network_cfg)


def load_program(name: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), name)
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text, flavour=NVFlavour()).parse()

    return program


def create_batch(
    program: QoalaProgram,
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
    deadline: int,
) -> BatchInfo:
    return BatchInfo(
        program=program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=deadline,
    )


def run(path: str) -> BatchResult:
    ns.sim_reset()

    node_config = get_config()
    network = create_network(node_config)
    procnode = network.nodes["alice"]

    num_iterations = 100
    inputs = [ProgramInput({}) for i in range(num_iterations)]

    unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())

    program = load_program(path)
    batch_info = create_batch(
        program=program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )

    procnode.submit_batch(batch_info)
    procnode.initialize_processes(linear=True)

    network.start_all_nodes()
    ns.sim_run()

    all_results = procnode.scheduler.get_batch_results()
    return all_results[0]


def test_1node_1qubit():
    batch_result = run("nv_1node_1qubit.iqoala")
    results = [result.values["m"] for result in batch_result.results]
    assert all(r == 1 for r in results)


def test_1node_2qubits_sg():
    batch_result = run("nv_1node_2qubits_sg.iqoala")
    results = [result.values["m"] for result in batch_result.results]
    assert all(r == [1, 1] for r in results)


def test_1node_2qubits_mg():
    batch_result = run("nv_1node_2qubits_mg.iqoala")
    results = [result.values["m"] for result in batch_result.results]
    assert all(r == [0, 0] for r in results)


def test_1node_5qubits():
    batch_result = run("nv_1node_5qubits.iqoala")
    results = [result.values["m"] for result in batch_result.results]
    assert all(r == [0, 0, 1, 0, 0] for r in results)


if __name__ == "__main__":
    # LogManager.set_log_level("DEBUG")
    test_1node_1qubit()
    test_1node_2qubits_sg()
    test_1node_2qubits_mg()
    test_1node_5qubits()



from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

from netqasm.lang.instr.flavour import Flavour, VanillaFlavour
from netqasm.lang.operand import Template
from netqasm.lang.parsing.text import parse_text_subroutine

from qoala.lang import hostlang as hl
from qoala.lang.hostlang import IqoalaSingleton, IqoalaValue, IqoalaVar, IqoalaVector
from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import RoutineMetadata

LHR_OP_NAMES: Dict[str, hl.ClassicalIqoalaOp] = {
    cls.OP_NAME: cls  # type: ignore
    for cls in [
        hl.SendCMsgOp,
        hl.ReceiveCMsgOp,
        hl.AddCValueOp,
        hl.MultiplyConstantCValueOp,
        hl.BitConditionalMultiplyConstantCValueOp,
        hl.AssignCValueOp,
        hl.RunSubroutineOp,
        hl.RunRequestOp,
        hl.ReturnResultOp,
        hl.BusyOp,
        hl.JumpOp,
        hl.BranchIfEqualOp,
        hl.BranchIfNotEqualOp,
        hl.BranchIfLessThanOp,
        hl.BranchIfGreaterThanOp,
    ]
}


class EndOfTextException(Exception):
    pass


class QoalaParseError(Exception):
    pass


def is_valid_name(name: str) -> bool:
    """Check if a string is a valid variable name.

    :param name: Name to check.

    :return: True if the string is a valid name, False otherwise. Acceptable names can only contain
    letters(a-z, A-Z), numbers(0-9) or underscores(_). Name must start with a letter, and cannot be
    a reserved keyword.

    """
    return name.isidentifier() and name[0].isalpha() and name not in LHR_OP_NAMES.keys()


class IqoalaMetaParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_meta_line(self, key: str, line: str) -> List[str]:

        if line.count(":") > 1:
            raise QoalaParseError("Qoala Program Meta lines must have a single colon.")
        split = line.split(":")

        if split[0] != key:
            raise QoalaParseError(
                f"Qoala Program Meta line expected to start with {key}, but found {split[0]}."
            )
        if len(split) == 1 or len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_meta_mapping(self, values: List[str]) -> Dict[int, str]:
        result_dict = {}
        for v in values:
            if v.count("->") != 1:
                raise QoalaParseError(
                    "Qoala Program Meta mapping must have a single arrow '->' between socket ID and "
                    "remote node name."
                )
            key_value = [x.strip() for x in v.split("->")]
            try:
                socket_id = int(key_value[0])
            except ValueError:
                raise QoalaParseError(
                    "Qoala Program Meta socket ID must be an integer."
                )
            result_dict[socket_id] = key_value[1]
        return result_dict

    def parse(self) -> ProgramMeta:
        try:
            start_line = self._read_line()
            if start_line != "META_START":
                raise QoalaParseError("Qoala Program Meta must start with META_START.")

            name_values = self._parse_meta_line("name", self._read_line())
            if len(name_values) != 1:
                raise QoalaParseError(
                    "Qoala Program Meta name must have a single value."
                )
            name = name_values[0].strip()
            if not is_valid_name(name):
                raise QoalaParseError(
                    f"Value {name} in Qoala Program Meta is not a valid program name."
                )

            parameters = self._parse_meta_line("parameters", self._read_line())
            for param in parameters:
                if not is_valid_name(param):
                    raise QoalaParseError(
                        f"Value {param} in Qoala Program Meta is not a valid parameter name."
                    )

            csockets_map = self._parse_meta_line("csockets", self._read_line())
            csockets = self._parse_meta_mapping(csockets_map)
            for node_name in csockets.values():
                if not is_valid_name(node_name):
                    raise QoalaParseError(
                        f"Value {node_name} in Qoala Program Meta is not a valid remote node name."
                    )
            epr_sockets_map = self._parse_meta_line("epr_sockets", self._read_line())
            epr_sockets = self._parse_meta_mapping(epr_sockets_map)
            for node_name in epr_sockets.values():
                if not is_valid_name(node_name):
                    raise QoalaParseError(
                        f"Value {node_name} in Qoala Program Meta is not a valid remote node name."
                    )

            end_line = self._read_line()
            if end_line != "META_END":
                raise QoalaParseError("Qoala Program Meta must start with META_END.")
        except EndOfTextException:
            raise QoalaParseError("Qoala Program Meta finished unexpectedly.")

        return ProgramMeta(name, parameters, csockets, epr_sockets)


class IqoalaInstrParser:
    def __init__(
        self, text: str, defined_vectors: Dict[str, hl.IqoalaVector] = None
    ) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0
        self._defined_vectors: Dict[str, hl.IqoalaVector] = {}
        if defined_vectors is not None:
            self._defined_vectors = defined_vectors

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_var(self, var_str: str, is_result=False) -> IqoalaVar:
        if var_str.startswith("tuple"):
            if var_str[5] != "<" or not var_str.endswith(">"):
                raise QoalaParseError(
                    "Iqoala tuple must start with 'tuple<' and end with '>'."
                )
            if var_str.count("<") > 1 or var_str.count(">") > 1:
                raise QoalaParseError(
                    "Iqoala tuple must have a single '<' and a single '>'."
                )
            tup_values_str = var_str[6:-1]
            if len(tup_values_str) == 0:
                tup_values = []
            else:
                tup_values = [x.strip() for x in tup_values_str.split(";")]
            for tup_value in tup_values:
                if not is_valid_name(tup_value):
                    raise QoalaParseError(
                        f"Value {tup_value} is not a valid variable name."
                    )
            return hl.IqoalaTuple(tup_values)
        elif "<" in var_str:
            if not is_result:
                raise QoalaParseError(
                    "Iqoala vector size cannot be specified for non-result variables."
                )
            if var_str.startswith("<"):
                raise QoalaParseError("Iqoala vector must start with a name.")
            if not var_str.endswith(">"):
                raise QoalaParseError("Iqoala vector must end with a '>'.")
            if var_str.count("<") != 1 or var_str.count(">") != 1:
                raise QoalaParseError(
                    "Iqoala vector must have a single '<' and a single '>'."
                )
            vec_split = var_str.split("<")
            vec_name = vec_split[0]
            if not is_valid_name(vec_name):
                raise QoalaParseError(f"Value {vec_name} is not a valid variable name.")
            vec_size_str = vec_split[1][:-1]  # strip last ">"
            vec_size: IqoalaValue  # Template is not used insde the blocks
            try:
                vec_size = int(vec_size_str)
            except ValueError:
                if not is_valid_name(vec_size_str):
                    raise QoalaParseError(
                        f"Value {vec_size_str} is not a valid variable name."
                    )
                vec_size = vec_size_str
            return hl.IqoalaVector(vec_name, vec_size)
        elif "[" in var_str:
            if var_str.startswith("["):
                raise QoalaParseError("Iqoala vector indexing must start with a name.")
            if not var_str.endswith("]"):
                raise QoalaParseError("Iqoala vector indexing must end with a ']'.")
            if var_str.count("[") != 1 or var_str.count("]") != 1:
                raise QoalaParseError(
                    "Iqoala vector indexing must have a single '[' and a single ']'."
                )
            vec_split = var_str.split("[")
            vec_name = vec_split[0]
            if not is_valid_name(vec_name):
                raise QoalaParseError(f"Value {vec_name} is not a valid variable name.")
            if vec_name not in self._defined_vectors:
                raise QoalaParseError(
                    f"Iqoala vector {vec_name} must be defined before indexing."
                )
            index_str = vec_split[1][:-1]  # strip last "]"
            index: int
            try:
                index = int(index_str)
            except ValueError:
                raise QoalaParseError("Iqoala vector indexing must be an integer.")
            return hl.IqoalaVectorElement(vec_name, index)
        else:
            if not is_valid_name(var_str):
                raise QoalaParseError(f"Value {var_str} is not a valid variable name.")
            if var_str in self._defined_vectors:
                return self._defined_vectors[var_str]
            return IqoalaSingleton(var_str)

    def _parse_lhr(self) -> hl.ClassicalIqoalaOp:
        line = self._read_line()

        attr: Optional[hl.IqoalaValue]

        assign_parts = [x.strip() for x in line.split("=")]
        if len(assign_parts) > 2:
            raise QoalaParseError("Iqoala instruction can have at most one '='.")
        if len(assign_parts) == 1:
            value = assign_parts[0]
            result = None
        elif len(assign_parts) == 2:
            value = assign_parts[1]
            result = self._parse_var(assign_parts[0], is_result=True)
            if isinstance(result, hl.IqoalaVector):
                self._defined_vectors[result.name] = result
        value_parts = [x.strip() for x in value.split(":")]
        if len(value_parts) > 2:
            raise QoalaParseError("Iqoala instruction can have at most one ':'.")
        if len(value_parts) == 2:
            value = value_parts[0].strip()
            attr_str = value_parts[1].strip()
            try:
                attr = int(attr_str)
            except ValueError:
                attr = attr_str
        else:
            value = value_parts[0]
            attr = None

        op_parts = [x.strip() for x in value.split("(")]
        if value.startswith("("):
            raise QoalaParseError("Iqoala instruction must have a name before '('.")
        if value.count("(") != 1 or value.count(")") != 1:
            raise QoalaParseError(
                "Iqoala instruction must have a single '(' and a single ')'."
            )
        if not value.endswith(")"):
            raise QoalaParseError(
                "In Iqoala instruction, operation must end with ')' or it must have a ')' "
                "before attributes."
            )
        op = op_parts[0]
        arguments = op_parts[1].rstrip(")")
        if len(arguments) == 0:
            raw_args = []
        else:
            raw_args = [x.strip() for x in arguments.split(",")]

        args = [self._parse_var(arg) for arg in raw_args]

        lhr_op = LHR_OP_NAMES[op].from_generic_args(result, args, attr)  # type: ignore
        return lhr_op

    def parse(self) -> List[hl.ClassicalIqoalaOp]:
        instructions: List[hl.ClassicalIqoalaOp] = []

        try:
            while True:
                instr = self._parse_lhr()
                instructions.append(instr)
        except EndOfTextException:
            pass

        return instructions


class HostCodeParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [
            line for line in lines if len(line) > 0 and not line.startswith("//")
        ]
        self._lineno: int = 0
        self._defined_vectors: Dict[str, hl.IqoalaVector] = {}  # name -> vector

    def get_block_texts(self) -> List[str]:
        block_start_lines: List[int] = []

        for i, line in enumerate(self._lines):
            if line.startswith("^"):
                block_start_lines.append(i)

        if len(block_start_lines) == 0:
            raise QoalaParseError("Qoala program must have at least one block.")

        block_texts: List[str] = []
        for i in range(len(block_start_lines) - 1):
            start = block_start_lines[i]
            end = block_start_lines[i + 1]
            text = self._lines[start:end]
            block_texts.append("\n".join([line for line in text]))

        last = block_start_lines[-1]
        last_text = self._lines[last:]
        block_texts.append("\n".join([line for line in last_text]))

        return block_texts

    def _parse_deadlines(self, text: str) -> Dict[str, int]:
        open_bracket = text.find("[")
        assert open_bracket >= 0
        close_bracket = text.find("]")
        assert close_bracket >= 0
        items = text[open_bracket + 1 : close_bracket].split(",")
        deadlines = {}
        for item in items:
            blk, dl = [i.strip() for i in item.split(":")]
            deadlines[blk] = int(dl)
        return deadlines

    def _parse_block_annotations(
        self, annotations: str
    ) -> Tuple[hl.BasicBlockType, Optional[Dict[str, int]]]:
        annotations_parts = annotations.split(",")
        if annotations_parts[0].count("=") != 1:
            raise QoalaParseError("Block type annotation must have exactly one '='.")
        type_annotation_parts = [x.strip() for x in annotations_parts[0].split("=")]
        if type_annotation_parts[0] != "type":
            raise QoalaParseError("Block type annotation must start with 'type'.")
        raw_typ = type_annotation_parts[1][-2:]
        try:
            typ = hl.BasicBlockType[raw_typ.upper()]
        except KeyError:
            raise QoalaParseError(
                "Invalid block type. Block type must be one of the "
                "following: 'CL', 'CC', 'QL', 'QC' (case insensitive)."
            )
        if len(annotations_parts) == 1:  # no deadline
            deadlines = None
        else:
            deadlines_str = annotations_parts[1].strip()
            deadlines = self._parse_deadlines(deadlines_str[12:])
        return typ, deadlines

    def _parse_block_header(
        self, line: str
    ) -> Tuple[str, hl.BasicBlockType, Optional[Dict[str, int]]]:
        # return (block name, block type, block deadline)
        if not line.startswith("^"):
            raise QoalaParseError("Block header must start with '^'.")

        if line.count("{") != 1 or line.count("}") != 1:
            raise QoalaParseError(
                "Block header must have a single '{' and a single '}'."
            )

        header_parts = [x.strip() for x in line.split("{")]

        name = header_parts[0][1:].strip()  # trim '^' at start
        if not is_valid_name(name):
            raise QoalaParseError(f"Value {name} is not a valid block name.")

        close_brace = header_parts[1].find("}")
        # Since we already checked that there is only one '{' and '}' in the header,
        # close_brace == -1 means '}' occured before '{'
        if close_brace == -1:
            raise QoalaParseError("'}' must occur after '{' in block header.")

        annotations_str = header_parts[1][:close_brace]
        if header_parts[1][close_brace + 1 :] != ":":
            raise QoalaParseError("Block header must end with ':'.")
        typ, deadline = self._parse_block_annotations(annotations_str)
        return name, typ, deadline

    def parse_block(self, text: str) -> hl.BasicBlock:
        lines = [line.strip() for line in text.split("\n")]
        lines = [line for line in lines if len(line) > 0]
        name, typ, deadline = self._parse_block_header(lines[0])
        instr_lines = lines[1:]
        instrs = IqoalaInstrParser(
            "\n".join(instr_lines), self._defined_vectors
        ).parse()

        return hl.BasicBlock(name, typ, instrs, deadline)

    def parse(self) -> List[hl.BasicBlock]:
        block_texts = self.get_block_texts()
        return [self.parse_block(text) for text in block_texts]


class LocalRoutineParser:
    def __init__(self, text: str, flavour: Optional[Flavour] = None) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0
        if flavour is None:
            flavour = VanillaFlavour()
        self._flavour = flavour

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_subrt_meta_line(self, key: str, line: str) -> List[str]:
        if line.count(":") != 1:
            raise QoalaParseError("SubRoutine Meta lines must have a single colon.")
        split = line.split(":")

        if split[0] != key:
            raise QoalaParseError(
                f"SubRoutine Meta line expected to start with {key}, but found {split[0]}."
            )

        if len(split) == 1:
            return []
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_subrt_meta_line_with_vecs(
        self, key: str, line: str
    ) -> List[Union[str, IqoalaVector]]:
        if line.count(":") != 1:
            raise QoalaParseError("SubRoutine Meta lines must have a single colon.")
        split = line.split(":")

        if split[0] != key:
            raise QoalaParseError(
                f"SubRoutine Meta line expected to start with {key}, but found {split[0]}."
            )
        if len(split) == 1:
            return []
        if len(split[1]) == 0:
            return []
        values_str = split[1].split(",")
        values_str = [v.strip() for v in values_str]
        values: List[Union[str, IqoalaVector]] = []
        for v in values_str:
            if "<" in v:
                if v.startswith("<"):
                    raise QoalaParseError("Iqoala vector must start with a name.")
                if not v.endswith(">"):
                    raise QoalaParseError("Iqoala vector must end with a '>'.")
                if v.count("<") != 1 or v.count(">") != 1:
                    raise QoalaParseError(
                        "Iqoala vector must have a single '<' and a single '>'."
                    )
                vec_split = v.split("<")
                vec_name = vec_split[0]
                if not is_valid_name(vec_name):
                    raise QoalaParseError(
                        f"Value {vec_name} is not a valid variable name."
                    )
                vec_size_str = vec_split[1][:-1].strip()  # strip last ">"
                vec_size: IqoalaValue
                if vec_size_str.startswith("{") and vec_size_str.endswith("}"):
                    vec_size_str = vec_size_str.strip("{}").strip()
                    if not is_valid_name(vec_size_str):
                        raise QoalaParseError(
                            f"Value {vec_size_str} is not a valid name for a template. "
                        )

                    vec_size = Template(vec_size_str)
                else:
                    try:
                        vec_size = int(vec_size_str)
                    except ValueError:
                        raise QoalaParseError(
                            "An integer or a template(using '{' and '{') must be provided "
                            f"as the size of an Iqoala vector {vec_name}."
                        )
                values.append(IqoalaVector(vec_name, vec_size))
            else:
                if not is_valid_name(v):
                    raise QoalaParseError(f"Value {v} is not a valid variable name.")
                values.append(v)
        return values

    def _parse_subroutine(self) -> LocalRoutine:
        name_line = self._read_line()
        if not name_line.startswith("SUBROUTINE "):
            raise QoalaParseError("SubRoutine Meta must start with 'SUBROUTINE'.")
        name = name_line[len("SUBROUTINE") + 1 :].strip()
        if not is_valid_name(name):
            raise QoalaParseError(f"Value {name} is not a valid SubRoutine name.")
        params_line = self._parse_subrt_meta_line("params", self._read_line())
        for param in params_line:
            if not is_valid_name(param):
                raise QoalaParseError(
                    f"SubRoutine {name}: value {param} is not a valid name for a parameter."
                )
        # TODO: use params line?

        return_vars = self._parse_subrt_meta_line_with_vecs(
            "returns", self._read_line()
        )
        assert all(" " not in v for v in return_vars if isinstance(v, str))

        uses_line = self._parse_subrt_meta_line("uses", self._read_line())
        try:
            uses = [int(u) for u in uses_line]
        except ValueError:
            raise QoalaParseError(
                f"SubRoutine {name}: 'uses' line values must be a comma separated list of integers."
            )
        keeps_line = self._parse_subrt_meta_line("keeps", self._read_line())
        try:
            keeps = [int(k) for k in keeps_line]
        except ValueError:
            raise QoalaParseError(
                f"SubRoutine {name}: 'keeps' line values must be a comma separated list of integers."
            )
        metadata = RoutineMetadata(qubit_use=uses, qubit_keep=keeps)

        request_line = self._parse_subrt_meta_line("request", self._read_line())
        if len(request_line) > 1:
            raise QoalaParseError(
                "SubRoutine request can have have at most one request."
            )
        request_name = None if len(request_line) == 0 else request_line[0]
        if request_name is not None and not is_valid_name(request_name):
            raise QoalaParseError(
                f"SubRoutine {name}: value {request_name} is not a valid name for a request."
            )

        start_line = self._read_line()
        if start_line != "NETQASM_START":
            raise QoalaParseError("SubRoutine must start with 'NETQASM_START'.")
        subrt_lines = []
        while True:
            try:
                line = self._read_line()
            except EndOfTextException:
                raise QoalaParseError("SubRoutine must end with 'NETQASM_END'.")
            if line == "NETQASM_END":
                break
            subrt_lines.append(line)
        subrt_text = "\n".join(subrt_lines)

        subrt = parse_text_subroutine(subrt_text, flavour=self._flavour)

        # Check that all templates are declared as params to the subroutine
        if any(arg not in params_line for arg in subrt.arguments):
            raise QoalaParseError(
                "All SubRoutine arguments must be declared in 'params'."
            )
        return LocalRoutine(name, subrt, return_vars, metadata, request_name)

    def parse(self) -> Dict[str, LocalRoutine]:
        subroutines: Dict[str, LocalRoutine] = {}
        try:
            while True:
                try:
                    subrt = self._parse_subroutine()
                    subroutines[subrt.name] = subrt
                except AssertionError:
                    raise QoalaParseError
        except EndOfTextException:
            return subroutines


class RequestRoutineParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_request_line(self, key: str, line: str) -> List[str]:
        if line.count(":") != 1:
            raise QoalaParseError("SubRoutine Meta lines must have a single colon.")
        split = line.split(":")

        if split[0] != key:
            raise QoalaParseError(
                f"SubRoutine Meta line expected to start with {key}, but found {split[0]}."
            )
        if len(split) == 1:
            return []
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_request_line_with_vecs(
        self, key: str, line: str
    ) -> List[Union[str, IqoalaVector]]:
        if line.count(":") != 1:
            raise QoalaParseError("SubRoutine Meta lines must have a single colon.")
        split = line.split(":")

        if split[0] != key:
            raise QoalaParseError(
                f"SubRoutine Meta line expected to start with {key}, but found {split[0]}."
            )
        if len(split) == 1:
            return []
        if len(split[1]) == 0:
            return []
        values_str = split[1].split(",")
        values_str = [v.strip() for v in values_str]
        values: List[Union[str, IqoalaVector]] = []
        for v in values_str:
            if "<" in v:
                if v.startswith("<"):
                    raise QoalaParseError("Iqoala vector must start with a name.")
                if not v.endswith(">"):
                    raise QoalaParseError("Iqoala vector must end with a '>'.")
                if v.count("<") != 1 or v.count(">") != 1:
                    raise QoalaParseError(
                        "Iqoala vector must have a single '<' and a single '>'."
                    )
                vec_split = v.split("<")
                vec_name = vec_split[0]
                if not is_valid_name(vec_name):
                    raise QoalaParseError(
                        f"Value {vec_name} is not a valid variable name."
                    )
                vec_size_str = vec_split[1][:-1].strip()  # strip last ">"
                vec_size: IqoalaValue
                if vec_size_str.startswith("{") and vec_size_str.endswith("}"):
                    vec_size_str = vec_size_str.strip("{}").strip()
                    if not is_valid_name(vec_size_str):
                        raise QoalaParseError(
                            f"Value {vec_size_str} is not a valid name for a template. "
                        )

                    vec_size = Template(vec_size_str)
                else:
                    try:
                        vec_size = int(vec_size_str)
                    except ValueError:
                        raise QoalaParseError(
                            "An integer or a template(using '{' and '{') must be provided "
                            f"as the size of an Iqoala vector {vec_name}."
                        )
                values.append(IqoalaVector(vec_name, vec_size))
            else:
                if not is_valid_name(v):
                    raise QoalaParseError(f"Value {v} is not a valid variable name.")
                values.append(v)
        return values

    def _parse_single_int_value(self, key: str, line: str) -> Union[int, Template]:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError(
                f"One single integer or a template value is allowed in Request Routine Meta line for {key}."
            )
        value = strings[0]
        if value.startswith("{") and value.endswith("}"):
            value = value.strip("{}").strip()
            if not is_valid_name(value):
                raise QoalaParseError(
                    f"Value {value} is not a valid name for a template."
                )
            return Template(value)
        try:
            return int(value)
        except ValueError:
            raise QoalaParseError(f"Value for {key} must be an integer or a template.")

    def _parse_optional_str_value(self, key: str, line: str) -> Optional[str]:
        strings = self._parse_request_line(key, line)
        if len(strings) == 0:
            return None
        if len(strings) != 1:
            raise QoalaParseError(
                f"At most one value is allowed in Request Routine Meta line for {key}."
            )
        return strings[0]

    def _parse_single_float_value(self, key: str, line: str) -> Union[float, Template]:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError(
                f"One single floating point value is allowed in Request Routine Meta line for {key}."
            )
        value = strings[0]
        if value.startswith("{") and value.endswith("}"):
            value = value.strip("{}").strip()
            if not is_valid_name(value):
                raise QoalaParseError(
                    f"Value {value} is not a valid name for a template."
                )
            return Template(value)
        try:
            return float(value)
        except ValueError:
            raise QoalaParseError(f"Value for {key} must be a float or a template.")

    def _parse_epr_create_role_value(self, key: str, line: str) -> EprRole:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError(
                "There must be a single value in Request Routine Meta line for EPR Role."
            )
        try:
            return EprRole[strings[0].upper()]
        except KeyError:
            raise QoalaParseError(
                "Value in Request Routine Meta line for EPR Role must be either"
                " 'CREATE' or 'RECEIVE' "
                "(case insensitive)."
            )

    def _parse_epr_create_type_value(self, key: str, line: str) -> EprType:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError(
                "There must be a single value in Request Routine Meta line for EPR Type."
            )
        try:
            return EprType[strings[0].upper()]
        except KeyError:
            raise QoalaParseError(
                "Value in Request Routine Meta line for EPR Type must be one of the following:"
                " 'CREATE_KEEP', 'MEASURE_DIRECTLY', 'REMOTE_STATE_PREP' "
                "(case insensitive)."
            )

    def _parse_virt_ids(self, key: str, line: str) -> RequestVirtIdMapping:
        if line.count(":") != 1:
            raise QoalaParseError("SubRoutine Meta lines must have a single colon.")
        split = line.split(":")

        if split[0] != key:
            raise QoalaParseError(
                f"SubRoutine Meta line expected to start with {key}, but found {split[0]}."
            )
        return RequestVirtIdMapping.from_str(split[1].strip())

    def _parse_callback_type(self, key: str, line: str) -> Optional[CallbackType]:
        values = self._parse_request_line(key, line)
        if len(values) == 0:
            return None
        if len(values) != 1:
            raise QoalaParseError(
                "There must be a single value in Request Routine Meta line for Callback Type."
            )
        try:
            return CallbackType[values[0].upper()]
        except KeyError:
            raise QoalaParseError(
                "If specified, value in Request Routine Meta line for Callback Type "
                "must be one of the following:"
                " 'SEQUENTIAL', 'WAIT_ALL'"
                "(case insensitive)."
            )

    def _parse_request(self) -> RequestRoutine:
        name_line = self._read_line()
        if not name_line.startswith("REQUEST "):
            raise QoalaParseError("Request Routine Meta must start with 'REQUEST'.")
        name = name_line[len("REQUEST") + 1 :].strip()
        if not is_valid_name(name):
            raise QoalaParseError(f"Value {name} is not a valid Request Routine name.")

        callback_type = self._parse_callback_type("callback_type", self._read_line())

        callback = self._parse_optional_str_value("callback", self._read_line())
        if callback is not None and not is_valid_name(callback):
            raise QoalaParseError(
                f"Request {name}: Value {callback} is not a valid callback name."
            )

        if callback is not None and callback_type is None:
            raise QoalaParseError(
                "If callback is specified, callback_type must also be specified."
            )
        # If both callback and callback_type are not specified, set callback_type to WAIT_ALL
        if callback_type is None:
            callback_type = CallbackType.WAIT_ALL

        return_vars = self._parse_request_line_with_vecs(
            "return_vars", self._read_line()
        )
        assert all(" " not in v for v in return_vars if isinstance(v, str))

        remote_id = self._parse_single_int_value("remote_id", self._read_line())
        epr_socket_id = self._parse_single_int_value("epr_socket_id", self._read_line())
        num_pairs = self._parse_single_int_value("num_pairs", self._read_line())
        virt_ids = self._parse_virt_ids("virt_ids", self._read_line())
        timeout = self._parse_single_int_value("timeout", self._read_line())
        fidelity = self._parse_single_float_value("fidelity", self._read_line())
        typ = self._parse_epr_create_type_value("typ", self._read_line())
        role = self._parse_epr_create_role_value("role", self._read_line())

        request = QoalaRequest(
            name=name,
            remote_id=remote_id,
            epr_socket_id=epr_socket_id,
            num_pairs=num_pairs,
            virt_ids=virt_ids,
            timeout=timeout,
            fidelity=fidelity,
            typ=typ,
            role=role,
        )
        return RequestRoutine(name, request, return_vars, callback_type, callback)

    def parse(self) -> Dict[str, RequestRoutine]:
        requests: Dict[str, RequestRoutine] = {}
        try:
            while True:
                request = self._parse_request()
                requests[request.name] = request
        except EndOfTextException:
            return requests


class QoalaParser:
    def __init__(
        self,
        text: Optional[str] = None,
        meta_text: Optional[str] = None,
        host_text: Optional[str] = None,
        subrt_text: Optional[str] = None,
        req_text: Optional[str] = None,
        flavour: Optional[Flavour] = None,
    ) -> None:
        if text is not None:
            if any(t is not None for t in (meta_text, host_text, subrt_text, req_text)):
                raise QoalaParseError(
                    "If text is provided to QoalaParser none of meta_text, "
                    "host_text, subrt_text, req_text must be provided."
                )
            meta_text, host_text, subrt_text, req_text = self._split_text(text)
        else:
            if (
                meta_text is None
                or host_text is None
                or subrt_text is None
                or req_text is None
            ):
                raise QoalaParseError(
                    "If text is not provided to QoalaParser, then all of meta_text, "
                    "host_text, subrt_text, req_text must be provided."
                )

        self._meta_text = meta_text
        self._host_text = host_text
        self._subrt_text = subrt_text
        self._req_text = req_text
        self._meta_parser = IqoalaMetaParser(meta_text)
        self._host_parser = HostCodeParser(host_text)
        self._subrt_parser = LocalRoutineParser(subrt_text, flavour)
        self._req_parser = RequestRoutineParser(req_text)

    def _split_text(self, text: str) -> Tuple[str, str, str, str]:
        lines = [line.strip() for line in text.split("\n")]
        meta_end_line: int
        first_subrt_line: Optional[int] = None
        first_req_line: Optional[int] = None
        for i, line in enumerate(lines):
            if "META_END" in line:
                meta_end_line = i
                break
        for i, line in enumerate(lines):
            if "SUBROUTINE" in line:
                first_subrt_line = i
                break
        for i, line in enumerate(lines):
            if "REQUEST" in line:
                first_req_line = i
                break

        meta_text = "\n".join(lines[0 : meta_end_line + 1])
        host_end_line: Optional[int] = None
        if first_subrt_line is None and first_req_line is None:
            # no subroutines and no requests
            subrt_text = ""
            req_text = ""
        elif first_subrt_line is not None and first_req_line is None:
            # subroutines but no requests
            subrt_text = "\n".join(lines[first_subrt_line:])
            req_text = ""
            host_end_line = first_subrt_line
        elif first_subrt_line is None and first_req_line is not None:
            # no subroutines but only requests
            subrt_text = ""
            req_text = "\n".join(lines[first_req_line:])
            host_end_line = first_req_line
        else:
            # subroutines and requests
            subrt_text = "\n".join(lines[first_subrt_line:first_req_line])
            req_text = "\n".join(lines[first_req_line:])
            host_end_line = first_subrt_line
        if host_end_line is not None:
            host_text = "\n".join(lines[meta_end_line + 1 : host_end_line])
        else:
            host_text = "\n".join(lines[meta_end_line + 1 :])

        return meta_text, host_text, subrt_text, req_text

    def parse(self) -> QoalaProgram:
        blocks = self._host_parser.parse()
        subroutines = self._subrt_parser.parse()
        requests = self._req_parser.parse()
        meta = self._meta_parser.parse()

        # Check that all references to subroutines (in RunSubroutineOp instructions)
        # and to requests (in RunRequestOp instructions) are valid.
        for block in blocks:
            for instr in block.instructions:
                if isinstance(instr, hl.RunSubroutineOp):
                    subrt_name = instr.subroutine
                    if subrt_name not in subroutines:
                        raise QoalaParseError(
                            f"Block {block.name} references unknown subroutine {subrt_name}"
                        )
                elif isinstance(instr, hl.RunRequestOp):
                    req_name = instr.req_routine
                    if req_name not in requests:
                        raise QoalaParseError(
                            f"Block {block.name} references unknown request routine {req_name}"
                        )

        return QoalaProgram(meta, blocks, subroutines, requests)



import itertools
from typing import List, Optional, Tuple

import netsquid as ns
from netsquid.nodes import Node

from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.lhi import LhiLinkInfo, LhiTopologyBuilder
from qoala.runtime.message import Message
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.entdist.entdist import EntDist, EntDistRequest
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.qdevice import QDevice
from qoala.util.math import B00_DENS, has_multi_state
from qoala.util.tests import netsquid_run


def create_n_qdevices(n: int, num_qubits: int = 1) -> List[QDevice]:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)
    qdevices: List[QDevice] = []
    for i in range(n):
        qproc = build_qprocessor_from_topology(name=f"qproc_{i}", topology=topology)
        node = Node(name=f"node_{i}", qmemory=qproc)
        qdevices.append(QDevice(node=node, topology=topology))

    return qdevices


def create_entdist(qdevices: List[QDevice]) -> EntDist:
    ehi_network = EhiNetworkInfo.only_nodes(
        {qdevice.node.ID: qdevice.node.name for qdevice in qdevices}
    )
    comp = EntDistComponent(ehi_network)
    entdist = EntDist(
        nodes=[qdevice.node for qdevice in qdevices], ehi_network=ehi_network, comp=comp
    )

    link_info = LhiLinkInfo.perfect(1000)
    for qd1, qd2 in itertools.combinations(qdevices, 2):
        entdist.add_sampler(qd1.node.ID, qd2.node.ID, link_info)

    return entdist


def create_request(
    node1_id: int,
    node2_id: int,
    local_qubit_id: int = 0,
    lpid: Optional[int] = 0,
    rpid: Optional[int] = 0,
) -> EntDistRequest:
    return EntDistRequest(
        local_node_id=node1_id,
        remote_node_id=node2_id,
        local_qubit_id=local_qubit_id,
        local_pid=lpid,
        remote_pid=rpid,
    )


def create_request_pair(
    node1_id: int,
    node2_id: int,
    node1_qubit_id: int = 0,
    node2_qubit_id: int = 0,
    lpid: Optional[int] = 0,
    rpid: Optional[int] = 0,
) -> Tuple[EntDistRequest]:
    req1 = EntDistRequest(
        local_node_id=node1_id,
        remote_node_id=node2_id,
        local_qubit_id=node1_qubit_id,
        local_pid=lpid,
        remote_pid=rpid,
    )
    req2 = EntDistRequest(
        local_node_id=node2_id,
        remote_node_id=node1_id,
        local_qubit_id=node2_qubit_id,
        local_pid=rpid,
        remote_pid=lpid,
    )
    return req1, req2


def test1():
    alice, bob = create_n_qdevices(2)
    entdist = create_entdist([alice, bob])

    request_alice = create_request(alice.node.ID, bob.node.ID, 0, [0], [0])
    request_bob = create_request(bob.node.ID, alice.node.ID, 0, [0], [0])

    entdist.put_request(request_alice)
    entdist.put_request(request_bob)

    ns.sim_reset()
    assert ns.sim_time() == 0
    netsquid_run(entdist.serve_all_requests())
    assert ns.sim_time() == 1000

    alice_qubit = alice.get_local_qubit(0)
    bob_qubit = bob.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)


def test2():
    qdevices = create_n_qdevices(4, num_qubits=2)
    assert all(qdevice.get_qubit_count() == 2 for qdevice in qdevices)

    entdist = create_entdist(qdevices)

    ids = [qdevices[i].node.ID for i in range(4)]

    req01, req10 = create_request_pair(ids[0], ids[1], 0, 0, [0], [0])
    entdist.put_request(req01)
    entdist.put_request(req10)

    req02, req20 = create_request_pair(ids[0], ids[2], 1, 0, [0], [0])
    entdist.put_request(req02)
    entdist.put_request(req20)

    req13, req31 = create_request_pair(ids[1], ids[3], 1, 0, [0], [0])
    entdist.put_request(req13)
    entdist.put_request(req31)

    ns.sim_reset()
    assert ns.sim_time() == 0
    netsquid_run(entdist.serve_all_requests())
    assert ns.sim_time() == 3 * 1000  # 3 request pairs

    n0_q0 = qdevices[0].get_local_qubit(0)
    n0_q1 = qdevices[0].get_local_qubit(1)
    n1_q0 = qdevices[1].get_local_qubit(0)
    n1_q1 = qdevices[1].get_local_qubit(1)
    n2_q0 = qdevices[2].get_local_qubit(0)
    n2_q1 = qdevices[2].get_local_qubit(1)
    n3_q0 = qdevices[3].get_local_qubit(0)
    n3_q1 = qdevices[3].get_local_qubit(1)

    assert has_multi_state([n0_q0, n1_q0], B00_DENS)
    assert has_multi_state([n0_q1, n2_q0], B00_DENS)
    assert has_multi_state([n1_q1, n3_q0], B00_DENS)

    assert n2_q1 is None
    assert n3_q1 is None


def test_run():
    qdevices = create_n_qdevices(4, num_qubits=2)
    assert all(qdevice.get_qubit_count() == 2 for qdevice in qdevices)

    entdist = create_entdist(qdevices)

    ids = [qdevices[i].node.ID for i in range(4)]

    req_01 = EntDistRequest(ids[0], ids[1], 0, [2], [0])
    req_10 = EntDistRequest(ids[1], ids[0], 0, [0], [2])
    req_02 = EntDistRequest(ids[0], ids[2], 1, [1], [1])
    req_20 = EntDistRequest(ids[2], ids[0], 0, [1], [1])
    req_13 = EntDistRequest(ids[1], ids[3], 1, [0], [1])
    req_31 = EntDistRequest(ids[3], ids[1], 1, [1], [0])

    ns.sim_reset()
    assert ns.sim_time() == 0

    node0, node1, node2, node3 = [qdevice.node for qdevice in qdevices]

    port0 = node0.add_ports("entdist_port")[0]
    port1 = node1.add_ports("entdist_port")[0]
    port2 = node2.add_ports("entdist_port")[0]
    port3 = node3.add_ports("entdist_port")[0]

    port0.connect(entdist._comp.node_in_port(node0.name))
    port1.connect(entdist._comp.node_in_port(node1.name))
    port2.connect(entdist._comp.node_in_port(node2.name))
    port3.connect(entdist._comp.node_in_port(node3.name))

    port0.tx_output(Message(-1, -1, req_01))
    port1.tx_output(Message(-1, -1, req_10))
    port0.tx_output(Message(-1, -1, req_02))
    port2.tx_output(Message(-1, -1, req_20))
    port1.tx_output(Message(-1, -1, req_13))
    port3.tx_output(Message(-1, -1, req_31))

    entdist.start()
    ns.sim_run()

    n0_q0 = qdevices[0].get_local_qubit(0)
    n0_q1 = qdevices[0].get_local_qubit(1)
    n1_q0 = qdevices[1].get_local_qubit(0)
    n1_q1 = qdevices[1].get_local_qubit(1)
    n2_q0 = qdevices[2].get_local_qubit(0)
    n2_q1 = qdevices[2].get_local_qubit(1)
    n3_q0 = qdevices[3].get_local_qubit(0)
    n3_q1 = qdevices[3].get_local_qubit(1)

    assert has_multi_state([n0_q0, n1_q0], B00_DENS)
    assert has_multi_state([n0_q1, n2_q0], B00_DENS)
    assert has_multi_state([n1_q1, n3_q1], B00_DENS)

    assert n2_q1 is None
    assert n3_q0 is None


if __name__ == "__main__":
    test1()
    test2()
    test_run()



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Dict, FrozenSet, Generator, List, Optional, Tuple

import netsquid as ns
from netsquid.nodes import Node
from netsquid.protocols import Protocol
from netsquid.qubits import qubitapi
from netsquid.qubits.qrepr import QRepr
from netsquid.qubits.qubit import Qubit
from netsquid_magic.state_delivery_sampler import (
    DeliverySample,
    IStateDeliverySamplerFactory,
    StateDeliverySampler,
)

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo, EhiNetworkTimebin
from qoala.runtime.lhi import LhiLinkInfo
from qoala.runtime.message import Message
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.entdist.entdistinterface import EntDistInterface
from qoala.sim.events import EPR_DELIVERY
from qoala.util.logging import LogManager


@dataclass(frozen=True)
class EntDistRequest:
    local_node_id: int
    remote_node_id: int
    local_qubit_id: int
    local_pid: int
    remote_pid: int

    def is_opposite(self, req: EntDistRequest) -> bool:
        return (
            self.local_node_id == req.remote_node_id
            and self.remote_node_id == req.local_node_id
            and self.local_pid == req.remote_pid
            and self.remote_pid == req.local_pid
        )

    def matches_timebin(self, bin: EhiNetworkTimebin) -> bool:
        if frozenset({self.local_node_id, self.remote_node_id}) != bin.nodes:
            return False
        return (
            bin.pids[self.local_node_id] == self.local_pid
            and bin.pids[self.remote_node_id] == self.remote_pid
        )


@dataclass(frozen=True)
class JointRequest:
    node1_id: int
    node2_id: int
    node1_qubit_id: int
    node2_qubit_id: int
    node1_pid: int
    node2_pid: int


@dataclass
class EprDeliverySample:
    state: QRepr
    duration: float

    @classmethod
    def from_ns_magic_delivery_sample(cls, sample: DeliverySample) -> EprDeliverySample:
        assert sample.state.num_qubits == 2
        return EprDeliverySample(
            state=sample.state.dm, duration=sample.delivery_duration
        )


@dataclass
class DelayedSampler:
    sampler: StateDeliverySampler
    delay: float


class EntDist(Protocol):
    def __init__(
        self,
        nodes: List[Node],
        ehi_network: EhiNetworkInfo,
        comp: EntDistComponent,
    ) -> None:
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._ehi_network = ehi_network
        self._comp = comp
        self._netschedule = ehi_network.network_schedule

        # Owned objects.
        self._interface = EntDistInterface(comp, ehi_network)

        # Node ID -> Node
        self._nodes: Dict[int, Node] = {node.ID: node for node in nodes}

        # (Node ID 1, Node ID 2) -> Sampler
        self._samplers: Dict[FrozenSet[int], DelayedSampler] = {}

        # Node ID -> list of requests
        self._requests: Dict[int, List[EntDistRequest]] = {
            node.ID: [] for node in nodes
        }

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}(EntDist)"
        )

    @property
    def comp(self) -> EntDistComponent:
        return self._comp

    def clear_requests(self) -> None:
        self._requests = {id: [] for id in self._nodes.keys()}

    def _add_sampler(
        self,
        node1_id: int,
        node2_id: int,
        factory: IStateDeliverySamplerFactory,
        kwargs: Dict[str, Any],
        delay: float,
    ) -> None:
        if node1_id == node2_id:
            raise ValueError("Cannot add sampler for same node.")
        if node1_id not in self._nodes:
            raise ValueError(f"Node {node1_id} not in network.")
        if node2_id not in self._nodes:
            raise ValueError(f"Node {node2_id} not in network.")

        link = frozenset([node1_id, node2_id])
        if link in self._samplers:
            raise ValueError(
                f"Sampler for ({node1_id}, {node2_id}) already registered \
                NOTE: only one sampler per node pair is allowed; order does not matter."
            )
        sampler = factory.create_state_delivery_sampler(**kwargs)
        self._samplers[link] = DelayedSampler(sampler, delay)

    def add_sampler(self, node1_id: int, node2_id: int, info: LhiLinkInfo) -> None:
        self._add_sampler(
            node1_id=node1_id,
            node2_id=node2_id,
            factory=info.sampler_factory(),
            kwargs=info.sampler_kwargs,
            delay=info.state_delay,
        )

    def get_sampler(self, node1_id: int, node2_id: int) -> DelayedSampler:
        link = frozenset([node1_id, node2_id])
        try:
            return self._samplers[link]
        except KeyError:
            raise ValueError(
                f"No sampler registered for pair ({node1_id}, {node2_id}) \
                NOTE: only one sampler per node pair is allowed; order does not matter."
            )

    def sample_state(cls, sampler: StateDeliverySampler) -> EprDeliverySample:
        raw_sample: DeliverySample = sampler.sample()
        return EprDeliverySample.from_ns_magic_delivery_sample(raw_sample)

    def create_epr_pair_with_state(cls, state: QRepr) -> Tuple[Qubit, Qubit]:
        q0, q1 = qubitapi.create_qubits(2)
        qubitapi.assign_qstate([q0, q1], state)
        return q0, q1

    def deliver(
        self,
        node1_id: int,
        node1_phys_id: int,
        node2_id: int,
        node2_phys_id: int,
        node1_pid: int,
        node2_pid: int,
    ) -> Generator[EventExpression, None, None]:
        timed_sampler = self.get_sampler(node1_id, node2_id)
        sample = self.sample_state(timed_sampler.sampler)
        epr = self.create_epr_pair_with_state(sample.state)

        self._logger.info(f"sample duration: {sample.duration}")
        self._logger.info(f"total duration: {timed_sampler.delay}")
        total_delay = sample.duration + timed_sampler.delay

        node1_mem = self._nodes[node1_id].qmemory
        node2_mem = self._nodes[node2_id].qmemory

        if not (0 <= node1_phys_id < node1_mem.num_positions):
            raise ValueError(
                f"qubit location id of {node1_phys_id} is not present in \
                    quantum memory of node ID {node1_id}."
            )
        if not (0 <= node2_phys_id < node2_mem.num_positions):
            raise ValueError(
                f"qubit location id of {node2_phys_id} is not present in \
                    quantum memory of node ID {node2_id}."
            )
        node1_mem.mem_positions[node1_phys_id].in_use = True
        node2_mem.mem_positions[node2_phys_id].in_use = True

        self._schedule_after(total_delay, EPR_DELIVERY)
        event_expr = EventExpression(source=self, event_type=EPR_DELIVERY)
        yield event_expr

        self._logger.info("pair delivered")

        node1_mem.put(qubits=epr[0], positions=node1_phys_id)
        node2_mem.put(qubits=epr[1], positions=node2_phys_id)

        # Send messages to the nodes indictating a request has been delivered.
        node1 = self._interface.remote_id_to_peer_name(node1_id)
        node2 = self._interface.remote_id_to_peer_name(node2_id)
        # TODO: use PIDs??
        self._interface.send_node_msg(node1, Message(-1, -1, node1_pid))
        self._interface.send_node_msg(node2, Message(-1, -1, node2_pid))

    def put_request(self, request: EntDistRequest) -> None:
        if request.local_node_id not in self._nodes:
            raise ValueError(
                f"Invalid request: node ID {request.local_node_id} not registered in EntDist."
            )
        if request.remote_node_id not in self._nodes:
            raise ValueError(
                f"Invalid request: node ID {request.remote_node_id} not registered in EntDist."
            )
        if request.remote_node_id == request.local_node_id:
            raise ValueError(
                f"Invalid request: local node ID {request.local_node_id} and remote node ID "
                "{request.remote_node_id} are the same."
            )

        self._requests[request.local_node_id].append(request)

    def get_requests(self, node_id: int) -> List[EntDistRequest]:
        return self._requests[node_id]

    def pop_request(self, node_id: int, index: int) -> EntDistRequest:
        return self._requests[node_id].pop(index)

    def get_remote_request_for(self, local_request: EntDistRequest) -> Optional[int]:
        """Return index in the request list of the remote node."""

        try:
            remote_requests = self._requests[local_request.remote_node_id]
        except KeyError:
            # Invalid remote node ID.
            raise ValueError(
                f"Request {local_request} refers to remote node \
                {local_request.remote_node_id} \
                but this node is not registed in the EntDist."
            )

        for i, req in enumerate(remote_requests):
            # Find the remote request that corresponds to the local request.
            if local_request.is_opposite(req):
                return i

        return None

    def get_next_joint_request(self) -> Optional[JointRequest]:
        for _, local_requests in self._requests.items():
            if len(local_requests) == 0:
                continue
            local_request = local_requests.pop(0)
            remote_request_id = self.get_remote_request_for(local_request)
            if remote_request_id is not None:
                remote_id = local_request.remote_node_id
                remote_request = self._requests[remote_id].pop(remote_request_id)
                return JointRequest(
                    local_request.local_node_id,
                    remote_request.local_node_id,
                    local_request.local_qubit_id,
                    remote_request.local_qubit_id,
                    local_request.local_pid,
                    local_request.remote_pid,
                )
            else:
                # Put local request back
                local_requests.insert(0, local_request)

        # No joint requests found
        return None

    def serve_request(
        self, request: JointRequest
    ) -> Generator[EventExpression, None, None]:
        yield from self.deliver(
            node1_id=request.node1_id,
            node1_phys_id=request.node1_qubit_id,
            node2_id=request.node2_id,
            node2_phys_id=request.node2_qubit_id,
            node1_pid=request.node1_pid,
            node2_pid=request.node2_pid,
        )

    def serve_all_requests(self) -> Generator[EventExpression, None, None]:
        while (request := self.get_next_joint_request()) is not None:
            yield from self.serve_request(request)

    def start(self) -> None:
        assert self._interface is not None
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()

    def _run_with_netschedule(self) -> Generator[EventExpression, None, None]:
        assert self._netschedule is not None

        while True:
            # Wait until a message arrives.
            yield from self._interface.wait_for_any_msg()

            # Wait until the next time bin.
            now = ns.sim_time()
            next_slot_time, next_slot = self._netschedule.next_bin(now)

            if next_slot_time - now > 0:
                yield from self._interface.wait(next_slot_time - now + 1)
            elif next_slot_time - now < 0:
                raise RuntimeError()

            messages = self._interface.pop_all_messages()

            requesting_nodes: List[int] = []
            for msg in messages:
                self._logger.info(f"received new msg from node: {msg}")
                request: EntDistRequest = msg.content
                requesting_nodes.append(request.local_node_id)

                self._logger.info(f"netschedule: {self._netschedule}")
                self._logger.info(f"next slot: {next_slot}")
                if request.matches_timebin(next_slot):
                    self._logger.info(f"putting request: {request}")
                    self.put_request(request)
                else:
                    self._logger.info(f"not handling msg {msg} (wrong timebin)")

            joint_request = self.get_next_joint_request()
            if joint_request is not None:
                self._logger.info(f"serving request {joint_request}")
                yield from self.serve_request(joint_request)
                self._logger.info("served request")
                while next_request := self.get_next_joint_request():
                    self._logger.info(f"serving request {next_request}")
                    yield from self.serve_request(next_request)
                    self._logger.info("served request")
            else:
                yield from self._interface.wait(1000)
                for node_id in requesting_nodes:
                    node = self._interface.remote_id_to_peer_name(node_id)
                    self._interface.send_node_msg(node, Message(-1, -1, None))
            self.clear_requests()
            yield from self._interface.wait(1)

    def _run_without_netschedule(self) -> Generator[EventExpression, None, None]:
        while True:
            # Wait for a new message.
            msg = yield from self._interface.receive_msg()
            self._logger.info(f"received new msg from node: {msg}")
            request: EntDistRequest = msg.content
            self.put_request(request)
            yield from self.serve_all_requests()

    def run(self) -> Generator[EventExpression, None, None]:
        if self._netschedule is None:
            yield from self._run_without_netschedule()
        else:
            yield from self._run_with_netschedule()



from typing import Dict, List

from qoala.runtime.task import QoalaTask


class SchedulerStatistics:
    def __init__(
        self,
        cpu_tasks_executed: Dict[int, QoalaTask],
        qpu_tasks_executed: Dict[int, QoalaTask],
        cpu_task_starts: Dict[int, float],
        qpu_task_starts: Dict[int, float],
        cpu_task_ends: Dict[int, float],
        qpu_task_ends: Dict[int, float],
    ) -> None:
        # task ID -> task
        self._cpu_tasks_executed = cpu_tasks_executed
        self._qpu_tasks_executed = qpu_tasks_executed

        # task ID -> start time
        self._cpu_task_starts = cpu_task_starts
        self._qpu_task_starts = qpu_task_starts

        # task ID -> end time
        self._cpu_task_ends: Dict[int, float] = cpu_task_ends
        self._qpu_task_ends: Dict[int, float] = qpu_task_ends

        cpu_pids = set([t.pid for t in self._cpu_tasks_executed.values()])
        qpu_pids = set([t.pid for t in self._qpu_tasks_executed.values()])
        pids = cpu_pids.union(qpu_pids)
        self._pids = list(pids)

        # for pid in self._pids

    @property
    def pids(self) -> List[int]:
        return self._pids

    @property
    def num_tasks_executed(self) -> int:
        return self.num_cpu_tasks_executed + self.num_qpu_tasks_executed

    @property
    def num_cpu_tasks_executed(self) -> int:
        return len(self._cpu_tasks_executed)

    @property
    def num_qpu_tasks_executed(self) -> int:
        return len(self._qpu_tasks_executed)

    def __str__(self) -> str:
        return (
            f"# tasks executed: {self.num_tasks_executed} "
            f"(CPU: {self.num_cpu_tasks_executed}, "
            f"QPU: {self.num_qpu_tasks_executed})"
            f"\npids: {self.pids}"
        )



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Dict, Generator

from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.lang.hostlang import BasicBlockType, RunRequestOp, RunSubroutineOp
from qoala.runtime.message import LrCallTuple, RrCallTuple
from qoala.runtime.task import (
    HostEventTask,
    HostLocalTask,
    LocalRoutineTask,
    MultiPairCallbackTask,
    MultiPairTask,
    PostCallTask,
    PreCallTask,
    QoalaTask,
    SinglePairCallbackTask,
    SinglePairTask,
)
from qoala.sim.host.hostprocessor import HostProcessor
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack.netstackprocessor import NetstackProcessor
from qoala.sim.process import QoalaProcess
from qoala.sim.qnos.qnosprocessor import QnosProcessor
from qoala.util.logging import LogManager


class SharedSchedulerMemory:
    def __init__(self) -> None:
        # Used to share information between related tasks:
        # specifically the lrcall/rrcall tuples that are shared between
        # precall, postcall, and pair/callback tasks
        # Values are *only* written by PreCall tasks.
        self._shared_lrcalls: Dict[int, LrCallTuple] = {}
        self._shared_rrcalls: Dict[int, RrCallTuple] = {}

    def write_shared_lrcall(self, ptr: int, lrcall: LrCallTuple) -> None:
        self._shared_lrcalls[ptr] = lrcall

    def write_shared_rrcall(self, ptr: int, rrcall: RrCallTuple) -> None:
        self._shared_rrcalls[ptr] = rrcall

    def read_shared_lrcall(self, ptr: int) -> LrCallTuple:
        return self._shared_lrcalls[ptr]

    def read_shared_rrcall(self, ptr: int) -> RrCallTuple:
        return self._shared_rrcalls[ptr]


class Driver(Protocol):
    def __init__(self, name: str, memory: SharedSchedulerMemory) -> None:
        super().__init__(name=name)

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({name})"
        )

        self._memory = memory

    @abstractmethod
    def handle_task(self, task: QoalaTask) -> Generator[EventExpression, None, bool]:
        raise NotImplementedError


class CpuDriver(Driver):
    def __init__(
        self,
        node_name: str,
        memory: SharedSchedulerMemory,
        hostprocessor: HostProcessor,
        memmgr: MemoryManager,
    ) -> None:
        super().__init__(name=f"{node_name}_cpu_driver", memory=memory)

        self._task_logger = LogManager.get_task_logger(f"{node_name}_CpuDriver")

        self._hostprocessor = hostprocessor
        self._memmgr = memmgr

    def _handle_precall_lr(self, task: PreCallTask) -> None:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunSubroutineOp)

        # Let Host setup shared memory.
        lrcall = self._hostprocessor.prepare_lr_call(process, instr)
        # Store the lrcall object in the shared ptr, so other tasks can use it
        self._memory.write_shared_lrcall(task.shared_ptr, lrcall)

    def _handle_postcall_lr(self, task: PostCallTask) -> None:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunSubroutineOp)

        # The corresponding PreCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        lrcall = self._memory.read_shared_lrcall(task.shared_ptr)
        self._hostprocessor.post_lr_call(process, instr, lrcall)

    def _handle_precall_rr(self, task: PreCallTask) -> None:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunRequestOp)

        # Let Host setup shared memory.
        rrcall = self._hostprocessor.prepare_rr_call(process, instr)
        # Store the lrcall object in the shared ptr, so other tasks can use it
        self._memory.write_shared_rrcall(task.shared_ptr, rrcall)

    def _handle_postcall_rr(self, task: PostCallTask) -> None:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunRequestOp)

        # The corresponding PreCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        rrcall = self._memory.read_shared_rrcall(task.shared_ptr)
        self._hostprocessor.post_rr_call(process, instr, rrcall)

    def handle_task(self, task: QoalaTask) -> Generator[EventExpression, None, bool]:
        if isinstance(task, HostLocalTask) or isinstance(task, HostEventTask):
            process = self._memmgr.get_process(task.pid)
            yield from self._hostprocessor.assign_block(process, task.block_name)
        elif isinstance(task, PreCallTask):
            process = self._memmgr.get_process(task.pid)
            block = process.program.get_block(task.block_name)
            if block.typ == BasicBlockType.QL:
                self._handle_precall_lr(task)
            else:
                assert block.typ == BasicBlockType.QC
                self._handle_precall_rr(task)
            # Simulate processing time of PreCallTask
            # TODO refactor
            latency = self._hostprocessor._latencies.host_instr_time
            yield from self._hostprocessor._interface.wait(latency)
        elif isinstance(task, PostCallTask):
            process = self._memmgr.get_process(task.pid)
            block = process.program.get_block(task.block_name)
            if block.typ == BasicBlockType.QL:
                self._handle_postcall_lr(task)
            else:
                assert block.typ == BasicBlockType.QC
                self._handle_postcall_rr(task)
            # Simulate processing time of PostCallTask
            # TODO refactor
            latency = self._hostprocessor._latencies.host_instr_time
            yield from self._hostprocessor._interface.wait(latency)
        else:
            raise NotImplementedError

        return True


class QpuDriver(Driver):
    def __init__(
        self,
        node_name: str,
        memory: SharedSchedulerMemory,
        qnosprocessor: QnosProcessor,
        netstackprocessor: NetstackProcessor,
        memmgr: MemoryManager,
    ) -> None:
        super().__init__(name=f"{node_name}_qpu_driver", memory=memory)

        self._task_logger = LogManager.get_task_logger(f"{node_name}_QpuDriver")

        self._qnosprocessor = qnosprocessor
        self._netstackprocessor = netstackprocessor
        self._memmgr = memmgr

    def allocate_qubits_for_routine(
        self, process: QoalaProcess, routine_name: str
    ) -> None:
        routine = process.get_local_routine(routine_name)
        for virt_id in routine.metadata.qubit_use:
            if self._memmgr.phys_id_for(process.pid, virt_id) is None:
                self._memmgr.allocate(process.pid, virt_id)

    def free_qubits_after_routine(
        self, process: QoalaProcess, routine_name: str
    ) -> None:
        routine = process.get_local_routine(routine_name)
        for virt_id in routine.metadata.qubit_use:
            if virt_id not in routine.metadata.qubit_keep:
                self._task_logger.debug(f"freeing qubit {virt_id}")
                self._memmgr.free(process.pid, virt_id)
                phys_qubits_in_use = [
                    i
                    for i, vmap in self._memmgr._physical_mapping.items()
                    if vmap is not None
                ]
                self._task_logger.debug(f"physical qubits in use: {phys_qubits_in_use}")

    def _handle_local_routine(
        self, task: LocalRoutineTask
    ) -> Generator[EventExpression, None, None]:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunSubroutineOp)

        # The corresponding LrCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        lrcall: LrCallTuple = self._memory.read_shared_lrcall(task.shared_ptr)

        # Allocate required qubits.
        self.allocate_qubits_for_routine(process, lrcall.routine_name)
        # Execute the routine on Qnos.
        yield from self._qnosprocessor.assign_local_routine(
            process, lrcall.routine_name, lrcall.input_addr, lrcall.result_addr
        )
        # Free qubits that do not need to be kept.
        self.free_qubits_after_routine(process, lrcall.routine_name)

    def _handle_multi_pair(
        self, task: MultiPairTask
    ) -> Generator[EventExpression, None, bool]:
        process = self._memmgr.get_process(task.pid)

        # The corresponding PreCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        rrcall: RrCallTuple = self._memory.read_shared_rrcall(task.shared_ptr)

        global_args = process.prog_instance.inputs.values
        self._netstackprocessor.instantiate_routine(process, rrcall, global_args)

        result = yield from self._netstackprocessor.handle_multi_pair(
            process, rrcall.routine_name
        )
        self._logger.info(f"Driver result: {result}")
        return result

    def _handle_multi_pair_callback(
        self, task: MultiPairCallbackTask
    ) -> Generator[EventExpression, None, None]:
        process = self._memmgr.get_process(task.pid)

        # The corresponding PreCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        rrcall: RrCallTuple = self._memory.read_shared_rrcall(task.shared_ptr)

        yield from self._netstackprocessor.handle_multi_pair_callback(
            process, rrcall.routine_name, self._qnosprocessor
        )

    def _handle_single_pair(
        self, task: SinglePairTask
    ) -> Generator[EventExpression, None, bool]:
        process = self._memmgr.get_process(task.pid)

        # The corresponding PreCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        rrcall: RrCallTuple = self._memory.read_shared_rrcall(task.shared_ptr)

        global_args = process.prog_instance.inputs.values
        self._netstackprocessor.instantiate_routine(process, rrcall, global_args)

        result = yield from self._netstackprocessor.handle_single_pair(
            process, rrcall.routine_name, task.pair_index
        )
        self._logger.info(f"Driver result: {result}")
        return result

    def _handle_single_pair_callback(
        self, task: SinglePairCallbackTask
    ) -> Generator[EventExpression, None, None]:
        process = self._memmgr.get_process(task.pid)

        # The corresponding PreCallTask must have executed, and it must have written
        # to the sharded scheduler memory.
        rrcall: RrCallTuple = self._memory.read_shared_rrcall(task.shared_ptr)

        yield from self._netstackprocessor.handle_single_pair_callback(
            process, rrcall.routine_name, self._qnosprocessor, task.pair_index
        )

    def handle_task(self, task: QoalaTask) -> Generator[EventExpression, None, bool]:
        if isinstance(task, LocalRoutineTask):
            yield from self._handle_local_routine(task)
        elif isinstance(task, MultiPairTask):
            result = yield from self._handle_multi_pair(task)
            return result
        elif isinstance(task, MultiPairCallbackTask):
            yield from self._handle_multi_pair_callback(task)
        elif isinstance(task, SinglePairTask):
            result = yield from self._handle_single_pair(task)
            return result
        elif isinstance(task, SinglePairCallbackTask):
            yield from self._handle_single_pair_callback(task)
        else:
            raise NotImplementedError
        return True



from typing import Generator

import netsquid as ns
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.sim.events import EVENT_WAIT


def yield_from(generator: Generator):
    try:
        while True:
            next(generator)
    except StopIteration as e:
        return e.value


class SimpleNetSquidProtocol(Protocol):
    def __init__(self, generator: Generator) -> None:
        super().__init__("test")
        self._gen = generator
        self.result = None

    def run(self) -> Generator[EventExpression, None, None]:
        self.result = yield from self._gen


def netsquid_run(generator: Generator):
    prot = SimpleNetSquidProtocol(generator)
    prot.start()
    ns.sim_run()
    return prot.result


def netsquid_wait(delta_time: int):
    prot = Protocol()
    prot._schedule_after(delta_time, EVENT_WAIT)
    prot.start()
    ns.sim_run()


def text_equal(text1, text2) -> bool:
    # allows whitespace differences
    lines1 = [line.strip() for line in text1.split("\n") if len(line) > 0]
    lines2 = [line.strip() for line in text2.split("\n") if len(line) > 0]
    for line1, line2 in zip(lines1, lines2):
        if line1 != line2:
            return False
    return True



import os

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.task import TaskGraphBuilder
from qoala.util.taskgraph import TaskGraphWriter


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def disabled_test1():
    program = load_program("test_callbacks_2_pairs.iqoala")
    # graph = TaskGraphBuilder.from_program(program, 0)
    graph = TaskGraphBuilder.from_program(program, 0)

    TaskGraphWriter(graph).draw("graph2.png")


if __name__ == "__main__":
    # test1()
    pass



from qoala.lang.hostlang import (
    AssignCValueOp,
    BasicBlock,
    BasicBlockType,
    IqoalaSingleton,
    ReturnResultOp,
)
from qoala.util.tests import text_equal


def test_basic_block():
    block = BasicBlock(
        name="b0",
        typ=BasicBlockType.CL,
        instructions=[
            AssignCValueOp(IqoalaSingleton("value"), 3),
            ReturnResultOp(IqoalaSingleton("value")),
        ],
    )

    expected = """
^b0 {type = CL}:
    value = assign_cval() : 3
    return_result(value)
    """

    assert text_equal(str(block), expected)


if __name__ == "__main__":
    test_basic_block()



from __future__ import annotations

import os
import random
from dataclasses import dataclass
from typing import List

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app, run_two_node_app_separate_inputs


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, determ: bool
) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=determ,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class TeleportResult:
    alice_results: BatchResult
    bob_results: BatchResult


def run_teleport(num_iterations: int, different_inputs: bool = False) -> TeleportResult:
    ns.sim_reset()

    num_qubits = 4
    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits, determ=True)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits, determ=True)

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, 1e9)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn]
    pattern = [(alice_id, i, bob_id, i) for i in range(num_iterations)]
    network_cfg.netschedule = NetworkScheduleConfig(
        bin_length=1_500, first_bin=0, bin_pattern=pattern, repeat_period=20_000
    )

    alice_program = load_program("teleport_alice.iqoala")
    bob_program = load_program("teleport_bob.iqoala")

    if different_inputs:
        alice_inputs: List[ProgramInput] = []
        bob_inputs: List[ProgramInput] = []
        for _ in range(num_iterations):
            basis = random.randint(0, 5)
            alice_inputs.append(ProgramInput({"bob_id": bob_id, "state": basis}))
            bob_inputs.append(ProgramInput({"alice_id": alice_id, "state": basis}))

        app_result = run_two_node_app_separate_inputs(
            num_iterations=num_iterations,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_inputs, "bob": bob_inputs},
            network_cfg=network_cfg,
        )
    else:
        # state = 5 -> teleport |1> state
        alice_input = ProgramInput({"bob_id": bob_id, "state": 5})
        # state = 4 -> measure in +Z, i.e. expect a "1" outcome
        bob_input = ProgramInput({"alice_id": alice_id, "state": 4})

        app_result = run_two_node_app(
            num_iterations=num_iterations,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_input, "bob": bob_input},
            network_cfg=network_cfg,
        )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return TeleportResult(alice_result, bob_result)


def test_teleport():
    # LogManager.set_log_level("DEBUG")
    # LogManager.set_task_log_level("DEBUG")
    # LogManager.log_to_file("teleport.log")
    # LogManager.log_tasks_to_file("teleport_tasks.log")
    num_iterations = 2

    result = run_teleport(num_iterations=num_iterations)

    program_results = result.bob_results.results
    outcomes = [result.values["outcome"] for result in program_results]
    print(outcomes)
    assert all(outcome == 1 for outcome in outcomes)


def test_teleport_different_inputs():
    # LogManager.set_task_log_level("DEBUG")
    num_iterations = 2

    result = run_teleport(num_iterations=num_iterations, different_inputs=True)

    program_results = result.bob_results.results
    outcomes = [result.values["outcome"] for result in program_results]
    print(outcomes)
    assert all(outcome == 0 for outcome in outcomes)


if __name__ == "__main__":
    test_teleport()
    test_teleport_different_inputs()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

from qoala.lang.hostlang import (
    AssignCValueOp,
    BasicBlock,
    BasicBlockType,
    ClassicalIqoalaOp,
    IqoalaSingleton,
    IqoalaTuple,
    IqoalaVector,
    ReturnResultOp,
    RunSubroutineOp,
)
from qoala.lang.request import RequestRoutine
from qoala.lang.routine import LocalRoutine


@dataclass
class ProgramMeta:
    name: str
    parameters: List[str]  # list of parameter names (all have type int)
    csockets: Dict[int, str]  # socket ID -> remote node name
    epr_sockets: Dict[int, str]  # socket ID -> remote node name

    @classmethod
    def empty(cls, name: str) -> ProgramMeta:
        return ProgramMeta(name=name, parameters=[], csockets={}, epr_sockets={})

    def serialize(self) -> str:
        s = "META_START"
        s += f"\nname: {self.name}"
        s += f"\nparameters: {', '.join(self.parameters)}"
        s += f"\ncsockets: {', '.join(f'{k} -> {v}' for k,v in self.csockets.items())}"
        s += f"\nepr_sockets: {', '.join(f'{k} -> {v}' for k,v in self.epr_sockets.items())}"
        s += "\nMETA_END"
        return s


class QoalaProgram:
    def __init__(
        self,
        meta: ProgramMeta,
        blocks: List[BasicBlock],
        local_routines: Optional[Dict[str, LocalRoutine]] = None,
        request_routines: Optional[Dict[str, RequestRoutine]] = None,
    ) -> None:
        self._meta: ProgramMeta = meta

        # List to keep order of blocks.
        self._blocks: List[BasicBlock] = blocks
        # Dict to easily find blocks.
        self._block_mapping: Dict[str, BasicBlock] = {blk.name: blk for blk in blocks}
        self._local_routines: Dict[str, LocalRoutine]
        self._request_routines: Dict[str, RequestRoutine]

        if local_routines is None:
            self._local_routines = {}
        else:
            self._local_routines = local_routines

        if request_routines is None:
            self._request_routines = {}
        else:
            self._request_routines = request_routines

    @property
    def meta(self) -> ProgramMeta:
        return self._meta

    @property
    def blocks(self) -> List[BasicBlock]:
        return self._blocks

    @blocks.setter
    def blocks(self, new_blocks: List[BasicBlock]) -> None:
        self._blocks = new_blocks

    def get_block(self, name: str) -> BasicBlock:
        return self._block_mapping[name]

    def get_block_id(self, name: str) -> int:
        return self._blocks.index(self._block_mapping[name])

    @property
    def instructions(self) -> List[ClassicalIqoalaOp]:
        instrs = []
        for b in self.blocks:
            instrs.extend(b.instructions)
        return instrs

    @property
    def local_routines(self) -> Dict[str, LocalRoutine]:
        return self._local_routines

    @local_routines.setter
    def local_routines(self, new_local_routines: Dict[str, LocalRoutine]) -> None:
        self._local_routines = new_local_routines

    @property
    def request_routines(self) -> Dict[str, RequestRoutine]:
        return self._request_routines

    @request_routines.setter
    def request_routines(self, new_routines: Dict[str, RequestRoutine]) -> None:
        self._request_routines = new_routines

    def __str__(self) -> str:
        return "\n".join("  " + str(i) for i in self.instructions)

    def serialize_meta(self) -> str:
        return self.meta.serialize()

    def serialize_block(self, block: BasicBlock) -> str:
        return str(block)

    def serialize_host_code(self) -> str:
        return "\n\n".join(self.serialize_block(b) for b in self.blocks)

    def serialize_subroutines(self) -> str:
        return "\n".join(s.serialize() for s in self.local_routines.values())

    def serialize(self) -> str:
        return (
            self.meta.serialize()
            + "\n"
            + self.serialize_host_code()
            + "\n"
            + self.serialize_subroutines()
        )


class QoalaProgramBuilder:
    @classmethod
    def single_routine(cls, routine: LocalRoutine, args: List[int]) -> QoalaProgram:
        meta = ProgramMeta.empty("name")
        prepare_args: List[ClassicalIqoalaOp] = []
        arg_singletons = [IqoalaSingleton(f"arg_{i}") for i in range(len(args))]

        for i in range(len(args)):
            prepare_args.append(AssignCValueOp(arg_singletons[i], args[i]))

        result_vec = IqoalaVector("result", routine.get_return_size())
        arg_names = [arg.name for arg in arg_singletons]
        args_tup = IqoalaTuple(arg_names)
        call_instr = RunSubroutineOp(
            result=result_vec, values=args_tup, subrt=routine.name
        )
        return_instr = ReturnResultOp(IqoalaSingleton("result"))

        instructions = prepare_args + [call_instr, return_instr]

        block = BasicBlock("b0", typ=BasicBlockType.QL, instructions=instructions)

        return QoalaProgram(
            meta=meta, blocks=[block], local_routines={routine.name: routine}
        )



from __future__ import annotations

from netsquid.protocols import Protocol

from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.ntf import GenericNtf, NtfInterface, NvNtf, TrappedIonNtf
from qoala.sim.memmgr import MemoryManager
from qoala.sim.qdevice import QDevice
from qoala.sim.qnos.qnoscomp import QnosComponent
from qoala.sim.qnos.qnosinterface import QnosInterface, QnosLatencies
from qoala.sim.qnos.qnosprocessor import (
    GenericProcessor,
    IonTrapProcessor,
    NVProcessor,
    QnosProcessor,
)


class Qnos(Protocol):
    """NetSquid protocol representing a QNodeOS instance."""

    def __init__(
        self,
        comp: QnosComponent,
        ehi_network: EhiNetworkInfo,
        memmgr: MemoryManager,
        qdevice: QDevice,
        latencies: QnosLatencies,
        ntf_interface: NtfInterface,
        asynchronous: bool = False,
    ) -> None:
        """Qnos protocol constructor.

        :param comp: NetSquid component representing the QNodeOS instance
        :param qdevice_type: hardware type of the QDevice of this node
        """
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._comp = comp
        self._ehi_network = ehi_network

        # Owned objects.
        self._interface = QnosInterface(comp, qdevice, memmgr)
        self._processor: QnosProcessor
        self._asynchronous = asynchronous

        self.create_processor(ntf_interface, latencies)

    def create_processor(
        self, ntf_interface: NtfInterface, latencies: QnosLatencies
    ) -> None:
        if isinstance(ntf_interface, GenericNtf):
            self._processor = GenericProcessor(
                self._interface, latencies, self._asynchronous
            )
        elif isinstance(ntf_interface, NvNtf):
            self._processor = NVProcessor(
                self._interface, latencies, self._asynchronous
            )
        elif isinstance(ntf_interface, TrappedIonNtf):
            self._processor = IonTrapProcessor(
                self._interface, latencies, self._asynchronous
            )
        else:
            raise ValueError

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    @qdevice.setter
    def qdevice(self, qdevice: QDevice) -> None:
        self._interface._qdevice = qdevice

    @property
    def processor(self) -> QnosProcessor:
        return self._processor

    @processor.setter
    def processor(self, processor: QnosProcessor) -> None:
        self._processor = processor

    def start(self) -> None:
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()
